/*
 * $RCSfile: AnalogInputSubsystem.cpp,v $
 * $Revision: 1.21 $
 * $Date: 2009/11/23 21:37:35 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class AnalogInputSubsystem implementation
 */


// {{{ includes
#include <assert.h>
#include <string.h>
#include "../lib/AIOUSB_Core.h"
#include "USBDeviceManager.hpp"
#include "AnalogInputSubsystem.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class AnalogInputSubsystem implementation

AnalogInputSubsystem::AnalogInputSubsystem( USBDevice &parent )
		: DeviceSubsystem( parent ) {
	autoCalFeature = AUTO_CAL_UNKNOWN;
	readBulkBuffer = 0;
	readBulkSamplesRequested = readBulkSamplesRetrieved = 0;
	DeviceProperties properties;
	int result = AIOUSB_GetDeviceProperties( getDeviceIndex(), &properties );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	numChannels = properties.ADCChannels;
	assert( numChannels > 0 );
	numMUXChannels = properties.ADCMUXChannels;
	ADConfigBlock temp;
	AIOUSB_InitConfigBlock( &temp, getDeviceIndex(), AIOUSB_FALSE );
	configBlock.resize( temp.size );
	validateConfigBlock();
	prevConfigBlock.resize( configBlock.size() );
	unsigned long size = configBlock.size();
	result = ADC_GetConfig( getDeviceIndex(), configBlock.data(), &size );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	prevConfigBlock = configBlock;
}	// AnalogInputSubsystem::AnalogInputSubsystem()


AnalogInputSubsystem::~AnalogInputSubsystem() {
	if( readBulkBuffer != 0 )
		delete[] readBulkBuffer;				// just in case ...
}	// AnalogInputSubsystem::~AnalogInputSubsystem()


void AnalogInputSubsystem::validateConfigBlock() const {
	assert( configBlock.size() == NUM_CONFIG_REGISTERS
		|| configBlock.size() == NUM_MUX_CONFIG_REGISTERS );
}	// AnalogInputSubsystem::validateConfigBlock()


AnalogInputSubsystem &AnalogInputSubsystem::writeConfig( bool force ) {
	validateConfigBlock();
	assert( prevConfigBlock.size() == NUM_CONFIG_REGISTERS
		|| prevConfigBlock.size() == NUM_MUX_CONFIG_REGISTERS );
	if(
		force
		|| prevConfigBlock != configBlock
	) {
		/*
		 * configuration block has been modified, send to device
		 */
		prevConfigBlock = configBlock;
		unsigned long size = prevConfigBlock.size();
		const int result = ADC_SetConfig( getDeviceIndex(), prevConfigBlock.data(), &size );
		if( result != AIOUSB_SUCCESS )
			throw OperationFailedException( result );
	}	// if( force ...
	return *this;
}	// AnalogInputSubsystem::writeConfig()


int AnalogInputSubsystem::getStartChannel() const {
	validateConfigBlock();
	int startChannel;
	if( configBlock.size() == NUM_MUX_CONFIG_REGISTERS )
		startChannel
			= ( ( configBlock.at( AD_CONFIG_MUX_START_END ) & 0x0f ) << 4 )
			| ( configBlock.at( AD_CONFIG_START_END ) & 0xf );
	else
		startChannel = configBlock.at( AD_CONFIG_START_END ) & 0xf;
	return startChannel;
}	// AnalogInputSubsystem::getStartChannel()


int AnalogInputSubsystem::getEndChannel() const {
	validateConfigBlock();
	int endChannel;
	if( configBlock.size() == NUM_MUX_CONFIG_REGISTERS )
		endChannel
			= ( configBlock.at( AD_CONFIG_MUX_START_END ) & 0xf0 )
			| ( ( configBlock.at( AD_CONFIG_START_END ) & 0xff ) >> 4 );
	else
		endChannel = ( configBlock.at( AD_CONFIG_START_END ) & 0xff ) >> 4;
	return endChannel;
}	// AnalogInputSubsystem::getEndChannel()


AnalogInputSubsystem &AnalogInputSubsystem::setScanRange( int startChannel, int numChannels ) {
	validateConfigBlock();
	if( configBlock.size() == NUM_MUX_CONFIG_REGISTERS ) {
		/*
		 * this board has a MUX, so support more channels
		 */
		if(
			numChannels < 1
			|| startChannel < 0
			|| startChannel + numChannels > numMUXChannels
		)
			throw IllegalArgumentException( "Invalid start channel or number of channels" );
		const int endChannel = startChannel + numChannels - 1;
		configBlock.at( AD_CONFIG_START_END ) = ( unsigned char ) ( ( endChannel << 4 ) | ( startChannel & 0x0f ) );
		configBlock.at( AD_CONFIG_MUX_START_END ) = ( unsigned char ) ( ( endChannel & 0xf0 ) | ( ( startChannel >> 4 ) & 0x0f ) );
	} else {
		/*
		 * this board doesn't have a MUX, so support base number of channels
		 */
		if(
			numChannels < 1
			|| startChannel < 0
			|| startChannel + numChannels > this->numChannels
		)
			throw IllegalArgumentException( "Invalid start channel or number of channels" );
		const int endChannel = startChannel + numChannels - 1;
		configBlock.at( AD_CONFIG_START_END ) = ( unsigned char ) ( ( endChannel << 4 ) | startChannel );
	}	// if( configBlock.size() ...
	return writeConfig( false );
}	// AnalogInputSubsystem::setScanRange()

/*
 * properties
 */

ostream &AnalogInputSubsystem::print( ostream &out ) {
	assert( &out != 0 );
	out
		<< "    Number of A/D channels: " << dec << numChannels << endl
		<< "    Number of MUXed A/D channels: " << numMUXChannels << endl;
	return out;
}	// AnalogInputSubsystem::print()


bool AnalogInputSubsystem::isAutoCalPresent( bool force ) {
	if(
		force
		|| autoCalFeature == AUTO_CAL_UNKNOWN
	) {
		const int result = ADC_QueryCal( getDeviceIndex() );
		if( result == AIOUSB_SUCCESS )
			autoCalFeature = AUTO_CAL_PRESENT;
		else if( result == AIOUSB_ERROR_NOT_SUPPORTED )
			autoCalFeature = AUTO_CAL_NOT_PRESENT;
		else {
			/*
			 * an error occurred; leave auto. cal. feature unchanged
			 */
			throw OperationFailedException( result );
		}	// else if( result ...
	}	// if( force ...
	return autoCalFeature == AUTO_CAL_PRESENT;
}	// AnalogInputSubsystem::isAutoCalPresent()

/*
 * configuration
 */

bool AnalogInputSubsystem::isDiscardFirstSample() const {
	return AIOUSB_IsDiscardFirstSample( getDeviceIndex() );
}	// AnalogInputSubsystem::isDiscardFirstSample()


AnalogInputSubsystem &AnalogInputSubsystem::setDiscardFirstSample( bool discard ) {
	const int result = AIOUSB_SetDiscardFirstSample( getDeviceIndex(), discard );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogInputSubsystem::setDiscardFirstSample()


int AnalogInputSubsystem::getCalMode() const {
	validateConfigBlock();
	return configBlock.at( AD_CONFIG_CAL_MODE );
}	// AnalogInputSubsystem::getCalMode()


AnalogInputSubsystem &AnalogInputSubsystem::setCalMode( int mode ) {
	validateConfigBlock();
	if(
		mode != AD_CAL_MODE_NORMAL
		&& mode != AD_CAL_MODE_GROUND
		&& mode != AD_CAL_MODE_REFERENCE
	)
		throw IllegalArgumentException( "Invalid cal. mode" );
	configBlock.at( AD_CONFIG_CAL_MODE ) = ( unsigned char ) mode;
	return writeConfig( false );
}	// AnalogInputSubsystem::setCalMode()


int AnalogInputSubsystem::getTriggerMode() const {
	validateConfigBlock();
	return configBlock.at( AD_CONFIG_TRIG_COUNT ) & AD_TRIGGER_VALID_MASK;
}	// AnalogInputSubsystem::getTriggerMode()


AnalogInputSubsystem &AnalogInputSubsystem::setTriggerMode( int mode ) {
	validateConfigBlock();
	if( ( mode & ~AD_TRIGGER_VALID_MASK ) != 0 )
		throw IllegalArgumentException( "Invalid trigger mode" );
	configBlock.at( AD_CONFIG_TRIG_COUNT ) = ( unsigned char ) mode;
	return writeConfig( false );
}	// AnalogInputSubsystem::setTriggerMode()


int AnalogInputSubsystem::getGainCode( int channel ) const {
	validateConfigBlock();
	if(
		channel < 0
		|| channel >= numMUXChannels
	)
		throw IllegalArgumentException( "Invalid channel" );
	return configBlock.at( AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ) & GAIN_CODE_VALID_MASK;
}	// AnalogInputSubsystem::getGainCode()


IntArray AnalogInputSubsystem::getGainCode( int startChannel, int numChannels ) const {
	validateConfigBlock();
	if(
		numChannels < 1
		|| startChannel < 0
		|| startChannel + numChannels > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid start channel or number of channels" );
	IntArray gainCode( numChannels );
	for( int index = 0; index < ( int ) gainCode.size(); index++ ) {
		gainCode.at( index )
			= configBlock.at( AD_CONFIG_GAIN_CODE + ( startChannel + index ) % AD_NUM_GAIN_CODE_REGISTERS )
			& GAIN_CODE_VALID_MASK;
	}	// for( int index ...
	return gainCode;
}	// AnalogInputSubsystem::getGainCode()


AnalogInputSubsystem &AnalogInputSubsystem::setGainCode( int channel, int gainCode ) {
	IntArray gainCodes( 1 );
	gainCodes.at( 0 ) = gainCode;
	return setGainCode( channel, gainCodes );
}	// AnalogInputSubsystem::setGainCode()


AnalogInputSubsystem &AnalogInputSubsystem::setGainCode( int startChannel, const IntArray &gainCode ) {
	validateConfigBlock();
	if(
		&gainCode == 0
		|| gainCode.size() < 1
		|| startChannel < 0
		|| startChannel + ( int ) gainCode.size() > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid gain code array or start channel" );
	for( int index = 0; index < ( int ) gainCode.size(); index++ ) {
		if( ( gainCode.at( index ) & ~GAIN_CODE_VALID_MASK ) != 0 )
			throw IllegalArgumentException( "Invalid gain code" );
		const int reg = AD_CONFIG_GAIN_CODE + ( startChannel + index ) % AD_NUM_GAIN_CODE_REGISTERS;
		configBlock.at( reg ) = ( unsigned char ) ( ( configBlock.at( reg ) & ~GAIN_CODE_VALID_MASK ) | gainCode.at( index ) );
	}	// for( int index ...
	return writeConfig( false );
}	// AnalogInputSubsystem::setGainCode()


bool AnalogInputSubsystem::isDifferentialMode( int channel ) const {
	validateConfigBlock();
	if(
		channel < 0
		|| channel >= numMUXChannels
	)
		throw IllegalArgumentException( "Invalid channel" );
	return ( configBlock.at( AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ) & AD_DIFFERENTIAL_MODE ) != 0;
}	// AnalogInputSubsystem::isDifferentialMode()


BoolArray AnalogInputSubsystem::isDifferentialMode( int startChannel, int numChannels ) const {
	validateConfigBlock();
	if(
		numChannels < 1
		|| startChannel < 0
		|| startChannel + numChannels > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid start channel or number of channels" );
	BoolArray differentialMode( numChannels );
	for( int index = 0; index < ( int ) differentialMode.size(); index++ ) {
		differentialMode.at( index )
			= ( configBlock.at( AD_CONFIG_GAIN_CODE + ( startChannel + index ) % AD_NUM_GAIN_CODE_REGISTERS ) & AD_DIFFERENTIAL_MODE ) != 0;
	}	// for( int index ...
	return differentialMode;
}	// AnalogInputSubsystem::isDifferentialMode()


AnalogInputSubsystem &AnalogInputSubsystem::setDifferentialMode( int channel, bool differentialMode ) {
	BoolArray differentialModes( 1 );
	differentialModes.at( 0 ) = differentialMode;
	return setDifferentialMode( channel, differentialModes );
}	// AnalogInputSubsystem::setDifferentialMode()


AnalogInputSubsystem &AnalogInputSubsystem::setDifferentialMode( int startChannel, const BoolArray &differentialMode ) {
	validateConfigBlock();
	if(
		&differentialMode == 0
		|| differentialMode.size() < 1
		|| startChannel < 0
		|| startChannel + ( int ) differentialMode.size() > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid differential mode array or start channel" );
	for( int index = 0; index < ( int ) differentialMode.size(); index++ ) {
		const int reg = AD_CONFIG_GAIN_CODE + ( startChannel + index ) % AD_NUM_GAIN_CODE_REGISTERS;
		if( differentialMode.at( index ) )
			configBlock.at( reg ) |= AD_DIFFERENTIAL_MODE;
		else
			configBlock.at( reg ) &= ~AD_DIFFERENTIAL_MODE;
	}	// for( int index ...
	return writeConfig( false );
}	// AnalogInputSubsystem::setDifferentialMode()


AnalogInputSubsystem &AnalogInputSubsystem::setGainCodeAndDiffMode( int channel, int gainCode, bool differentialMode ) {
	IntArray gainCodes( 1 );
	gainCodes.at( 0 ) = gainCode;
	BoolArray differentialModes( 1 );
	differentialModes.at( 0 ) = differentialMode;
	return setGainCodeAndDiffMode( channel, gainCodes, differentialModes );
}	// AnalogInputSubsystem::setGainCodeAndDiffMode()


AnalogInputSubsystem &AnalogInputSubsystem::setGainCodeAndDiffMode( int startChannel, const IntArray &gainCode, const BoolArray &differentialMode ) {
	validateConfigBlock();
	if(
		&gainCode == 0
		|| gainCode.size() < 1
		|| &differentialMode == 0
		|| differentialMode.size() != gainCode.size()
		|| startChannel < 0
		|| startChannel + ( int ) gainCode.size() > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid gain code array, differential mode array or start channel" );
	for( int index = 0; index < ( int ) gainCode.size(); index++ ) {
		if( ( gainCode.at( index ) & ~GAIN_CODE_VALID_MASK ) != 0 )
			throw IllegalArgumentException( "Invalid gain code" );
		const int reg = AD_CONFIG_GAIN_CODE + ( startChannel + index ) % AD_NUM_GAIN_CODE_REGISTERS;
		configBlock.at( reg ) = ( unsigned char ) ( ( configBlock.at( reg ) & ~GAIN_CODE_VALID_MASK ) | gainCode.at( index ) );
		if( differentialMode.at( index ) )
			configBlock.at( reg ) |= AD_DIFFERENTIAL_MODE;
		else
			configBlock.at( reg ) &= ~AD_DIFFERENTIAL_MODE;
	}	// for( int index ...
	return writeConfig( false );
}	// AnalogInputSubsystem::setGainCodeAndDiffMode()


AnalogInputSubsystem &AnalogInputSubsystem::setGainCodeAndDiffMode( int gainCode, bool differentialMode ) {
	validateConfigBlock();
	if( ( gainCode & ~GAIN_CODE_VALID_MASK ) != 0 )
		throw IllegalArgumentException( "Invalid gain code" );
	for( int index = 0; index < AD_NUM_GAIN_CODE_REGISTERS; index++ ) {
		const int reg = AD_CONFIG_GAIN_CODE + index;
		configBlock.at( reg ) = ( unsigned char ) ( ( configBlock.at( reg ) & ~GAIN_CODE_VALID_MASK ) | gainCode );
		if( differentialMode )
			configBlock.at( reg ) |= AD_DIFFERENTIAL_MODE;
		else
			configBlock.at( reg ) &= ~AD_DIFFERENTIAL_MODE;
	}	// for( int index ...
	return writeConfig( false );
}	// AnalogInputSubsystem::setGainCodeAndDiffMode()


int AnalogInputSubsystem::getOversample() const {
	validateConfigBlock();
	return configBlock.at( AD_CONFIG_OVERSAMPLE );
}	// AnalogInputSubsystem::getOversample()


AnalogInputSubsystem &AnalogInputSubsystem::setOversample( int oversample ) {
	validateConfigBlock();
	if(
		oversample < 0
		|| oversample > MAX_OVERSAMPLE
	)
		throw IllegalArgumentException( "Invalid oversample" );
	configBlock.at( AD_CONFIG_OVERSAMPLE ) = ( unsigned char ) oversample;
	return writeConfig( false );
}	// AnalogInputSubsystem::setOversample()


AnalogInputSubsystem &AnalogInputSubsystem::setCalibrationTable( const std::string &fileName ) {
	if(
		&fileName == 0
		|| fileName.empty()
	)
		throw IllegalArgumentException( "Invalid file name" );
	const int result = AIOUSB_ADC_LoadCalTable( getDeviceIndex(), fileName.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogInputSubsystem::setCalibrationTable()


AnalogInputSubsystem &AnalogInputSubsystem::setCalibrationTable( const UShortArray &calTable ) {
	if(
		&calTable == 0
		|| calTable.size() != CAL_TABLE_WORDS
	)
		throw IllegalArgumentException( "Invalid cal. table" );
	const int result = AIOUSB_ADC_SetCalTable( getDeviceIndex(), calTable.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogInputSubsystem::setCalibrationTable()


AnalogInputSubsystem &AnalogInputSubsystem::setStreamingBlockSize( int blockSize ) {
	if( blockSize < 0 )
		throw IllegalArgumentException( "Invalid block size" );
	const int result = AIOUSB_SetStreamingBlockSize( getDeviceIndex(), blockSize );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogInputSubsystem::setStreamingBlockSize()


AnalogInputSubsystem &AnalogInputSubsystem::setClock( double clockHz ) {
	parent->setMiscClock( clockHz );
	return *this;
}	// AnalogInputSubsystem::setClock()

/*
 * operations
 */

UShortArray AnalogInputSubsystem::calibrate( bool autoCal, bool returnCalTable, const std::string &saveFileName ) {
	UShortArray calTable( returnCalTable ? CAL_TABLE_WORDS : 0 );
	const int result = AIOUSB_ADC_InternalCal( getDeviceIndex(), autoCal
		, returnCalTable ? calTable.data() : 0
		, saveFileName.length() > 0 ? saveFileName.data() : 0 );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return calTable;
}	// AnalogInputSubsystem::calibrate()


UShortArray AnalogInputSubsystem::calibrate( const DoubleArray &points, bool returnCalTable, const std::string &saveFileName ) {
	if(
		&points == 0
		|| points.size() < 4					// at least 2 voltage-count pairs
		|| ( points.size() & 1 ) != 0
	)
		throw IllegalArgumentException( "Invalid points array" );
	UShortArray calTable( returnCalTable ? CAL_TABLE_WORDS : 0 );
	const int result = AIOUSB_ADC_ExternalCal( getDeviceIndex(), points.data(), points.size() / 2 /* voltage-count pairs */
		, returnCalTable ? calTable.data() : 0
		, saveFileName.length() > 0 ? saveFileName.data() : 0 );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return calTable;
}	// AnalogInputSubsystem::calibrate()


unsigned short AnalogInputSubsystem::read( int channel ) {
	return read( channel, 1 ).at( 0 );
}	// AnalogInputSubsystem::read()


UShortArray AnalogInputSubsystem::read( int startChannel, int numChannels ) {
	/*
	 * make copy of configuration block and restore it when done
	 */
	const UCharArray origConfigBlock( configBlock );

	/*
	 * if calibration mode is enabled, then don't permit more than one
	 * sample to be taken; otherwise "bad" things will happen
	 */
	int calMode = getCalMode();
	if(
		calMode == AD_CAL_MODE_GROUND
		|| calMode == AD_CAL_MODE_REFERENCE
	) {
		setDiscardFirstSample( false );
		setOversample( 0 );
		numChannels = 1;
	}	// if( calMode ...
	setScanRange( startChannel, numChannels );
	UShortArray counts( numChannels );
	const int result = AIOUSB_GetScan( getDeviceIndex(), counts.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );

	/*
	 * restore original configuration block
	 */
	configBlock = origConfigBlock;
	writeConfig( false );

	return counts;
}	// AnalogInputSubsystem::Read()


double AnalogInputSubsystem::readVolts( int channel ) {
	return readVolts( channel, 1 ).at( 0 );
}	// AnalogInputSubsystem::readVolts()


DoubleArray AnalogInputSubsystem::readVolts( int startChannel, int numChannels ) {
	return countsToVolts( startChannel, read( startChannel, numChannels ) );
}	// AnalogInputSubsystem::readVolts()


AnalogInputSubsystem &AnalogInputSubsystem::readBulkStart( int startChannel, int numChannels, int numSamples ) {
	if( numSamples < 1 )
		throw IllegalArgumentException( "Invalid number of samples" );
	readBulkBuffer = new unsigned short[ numSamples ];
	if( readBulkBuffer == 0 )
		throw OperationFailedException( AIOUSB_ERROR_NOT_ENOUGH_MEMORY );
	setScanRange( startChannel, numChannels );
	const int result = ADC_BulkAcquire( getDeviceIndex(), numSamples * sizeof( unsigned short ), readBulkBuffer );
	if( result == AIOUSB_SUCCESS ) {
		readBulkSamplesRequested = numSamples;
		readBulkSamplesRetrieved = 0;
	} else {
		delete[] readBulkBuffer;
		readBulkBuffer = 0;
		readBulkSamplesRequested = readBulkSamplesRetrieved = 0;
		throw OperationFailedException( result );
	}	// if( result ...
	return *this;
}	// AnalogInputSubsystem::readBulkStart()


int AnalogInputSubsystem::readBulkSamplesAvailable() {
	int samplesAvailable = 0;
	if( readBulkBuffer != 0 ) {
		unsigned long bytesRemaining;
		const int result = ADC_BulkPoll( getDeviceIndex(), &bytesRemaining );
		if( result != AIOUSB_SUCCESS )
			throw OperationFailedException( result );
		const int samplesRemaining = bytesRemaining / sizeof( unsigned short );
		samplesAvailable = readBulkSamplesRequested - readBulkSamplesRetrieved - samplesRemaining;
		if( samplesAvailable < 0 )
			samplesAvailable = 0;
	}	// if( readBulkBuffer ...
	return samplesAvailable;
}	// AnalogInputSubsystem::readBulkSamplesAvailable()


UShortArray AnalogInputSubsystem::readBulkNext( int numSamples ) {
	if( numSamples < 1 )
		throw IllegalArgumentException( "Invalid number of samples" );
	UShortArray counts;
	if( readBulkBuffer != 0 ) {
		int samplesAvailable = readBulkSamplesAvailable();
		if( samplesAvailable > 0 ) {
			if( numSamples > samplesAvailable )
				numSamples = samplesAvailable;
			counts.resize( numSamples );
			memcpy( counts.data(), readBulkBuffer + readBulkSamplesRetrieved, numSamples * sizeof( unsigned short ) );
			if( ( samplesAvailable -= numSamples ) <= 0 ) {
				/*
				 * no more data remains to be retrieved; finish bulk read procedure
				 */
				delete[] readBulkBuffer;
				readBulkBuffer = 0;
				readBulkSamplesRequested = readBulkSamplesRetrieved = 0;
			} else
				readBulkSamplesRetrieved += numSamples;
		}	// if( samplesAvailable ...
	}	// if( readBulkBuffer ...
	return counts;
}	// AnalogInputSubsystem::readBulkNext()

/*
 * utilities
 */

double AnalogInputSubsystem::countsToVolts( int channel, unsigned short counts ) const {
	if(
		channel < 0
		|| channel >= numMUXChannels
	)
		throw IllegalArgumentException( "Invalid channel" );
	return AIOUSB_CountsToVolts( getDeviceIndex(), channel, counts );
}	// AnalogInputSubsystem::countsToVolts()


DoubleArray AnalogInputSubsystem::countsToVolts( int startChannel, const UShortArray &counts ) const {
	if(
		&counts == 0
		|| counts.size() < 1
		|| startChannel < 0
		|| startChannel + ( int ) counts.size() > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid counts array or start channel" );
	DoubleArray volts( counts.size() );
	const int result = AIOUSB_ArrayCountsToVolts( getDeviceIndex(), startChannel, counts.size(), counts.data(), volts.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return volts;
}	// AnalogInputSubsystem::countsToVolts()


unsigned short AnalogInputSubsystem::voltsToCounts( int channel, double volts ) const {
	if(
		channel < 0
		|| channel >= numMUXChannels
	)
		throw IllegalArgumentException( "Invalid channel" );
	return AIOUSB_VoltsToCounts( getDeviceIndex(), channel, volts );
}	// AnalogInputSubsystem::voltsToCounts()


UShortArray AnalogInputSubsystem::voltsToCounts( int startChannel, const DoubleArray &volts ) const {
	if(
		&volts == 0
		|| volts.size() < 1
		|| startChannel < 0
		|| startChannel + ( int ) volts.size() > numMUXChannels
	)
		throw IllegalArgumentException( "Invalid volts array or start channel" );
	UShortArray counts( volts.size() );
	const int result = AIOUSB_ArrayVoltsToCounts( getDeviceIndex(), startChannel, volts.size(), volts.data(), counts.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return counts;
}	// AnalogInputSubsystem::voltsToCounts()

// }}}

}	// namespace AIOUSB


/* end of file */
