/*
 * $RCSfile: AnalogInputSubsystem.hpp,v $
 * $Revision: 1.17 $
 * $Date: 2009/11/23 21:37:35 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class AnalogInputSubsystem declarations
 */


#if ! defined( AnalogInputSubsystem_hpp )
#define AnalogInputSubsystem_hpp

// {{{ includes
#include <DeviceSubsystem.hpp>
// }}}

namespace AIOUSB {

// {{{ class AnalogInputSubsystem declarations

/*
 * class AnalogInputSubsystem represents the analog input subsystem of a board
 */

 class AnalogInputSubsystem : public DeviceSubsystem {
public:
	enum {
		MAX_COUNTS = 65535
	};	// enum

protected:
	enum {
		  NUM_CONFIG_REGISTERS			= 20	// number of "registers" (bytes) in config. block
		, NUM_MUX_CONFIG_REGISTERS		= 21	// number of "registers" (bytes) in config. block when MUX installed

		, GAIN_CODE_VALID_MASK			= 7
		, MAX_OVERSAMPLE				= 0xff

		, AUTO_CAL_UNKNOWN				= 0
		, AUTO_CAL_NOT_PRESENT			= 1
		, AUTO_CAL_PRESENT				= 2
	};	// enum

	int numChannels;							// number of A/D channels
	int numMUXChannels;							// number of MUXed A/D channels
	int autoCalFeature;							// cached A/D auto. cal. feature
	UCharArray configBlock;						// registers that comprise configuration block
	UCharArray prevConfigBlock;					// previous configuration block (so we can detect changes)
	unsigned short *readBulkBuffer;				// buffer used during bulk read procedure
	int readBulkSamplesRequested;				// total number of samples requested during bulk read procedure
	int readBulkSamplesRetrieved;				// number of samples retrieved so far during  bulk read procedure

	void validateConfigBlock() const;
	AnalogInputSubsystem &writeConfig( bool force );
	int getStartChannel() const;
	int getEndChannel() const;
	AnalogInputSubsystem &setScanRange( int startChannel, int numChannels );

public:
	AnalogInputSubsystem( USBDevice &parent );
	virtual ~AnalogInputSubsystem();

	// properties
	virtual std::ostream &print( std::ostream &out );

	int getNumChannels() const {
		return numChannels;
	}	// getNumChannels()

	int getNumMUXChannels() const {
		return numMUXChannels;
	}	// getNumMUXChannels()

	bool isAutoCalPresent( bool force );

	// configuration
	bool isDiscardFirstSample() const;
	AnalogInputSubsystem &setDiscardFirstSample( bool discard );
	int getCalMode() const;
	AnalogInputSubsystem &setCalMode( int mode );
	int getTriggerMode() const;
	AnalogInputSubsystem &setTriggerMode( int mode );
	int getGainCode( int channel ) const;
	IntArray getGainCode( int startChannel, int numChannels ) const;
	AnalogInputSubsystem &setGainCode( int channel, int gainCode );
	AnalogInputSubsystem &setGainCode( int startChannel, const IntArray &gainCode );
	bool isDifferentialMode( int channel ) const;
	BoolArray isDifferentialMode( int startChannel, int numChannels ) const;
	AnalogInputSubsystem &setDifferentialMode( int channel, bool differentialMode );
	AnalogInputSubsystem &setDifferentialMode( int startChannel, const BoolArray &differentialMode );
	AnalogInputSubsystem &setGainCodeAndDiffMode( int channel, int gainCode, bool differentialMode );
	AnalogInputSubsystem &setGainCodeAndDiffMode( int startChannel, const IntArray &gainCode, const BoolArray &differentialMode );
	AnalogInputSubsystem &setGainCodeAndDiffMode( int gainCode, bool differentialMode );
	int getOversample() const;
	AnalogInputSubsystem &setOversample( int oversample );
	AnalogInputSubsystem &setCalibrationTable( const std::string &fileName );
	AnalogInputSubsystem &setCalibrationTable( const UShortArray &calTable );
	AnalogInputSubsystem &setStreamingBlockSize( int blockSize );
	AnalogInputSubsystem &setClock( double clockHz );

	// operations
	UShortArray calibrate( bool autoCal, bool returnCalTable, const std::string &saveFileName );
	UShortArray calibrate( const DoubleArray &points, bool returnCalTable, const std::string &saveFileName );
	unsigned short read( int channel );
	UShortArray read( int startChannel, int numChannels );
	double readVolts( int channel );
	DoubleArray readVolts( int startChannel, int numChannels );
	AnalogInputSubsystem &readBulkStart( int startChannel, int numChannels, int numSamples );
	int readBulkSamplesAvailable();
	UShortArray readBulkNext( int numSamples );

	AnalogInputSubsystem &clearFIFO( int method ) {
		parent->clearFIFO( method );
		return *this;
	}	// clearFIFO()

	// utilities
	double countsToVolts( int channel, unsigned short counts ) const;
	DoubleArray countsToVolts( int startChannel, const UShortArray &counts ) const;
	unsigned short voltsToCounts( int channel, double volts ) const;
	UShortArray voltsToCounts( int startChannel, const DoubleArray &volts ) const;
};	// class AnalogInputSubsystem

// }}}

}	// namespace AIOUSB

#endif


/* end of file */
