/*
 * $RCSfile: AnalogOutputSubsystem.cpp,v $
 * $Revision: 1.8 $
 * $Date: 2009/11/29 17:21:05 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class AnalogOutputSubsystem implementation
 */


// {{{ includes
#include <assert.h>
#include <math.h>
#include <aiousb.h>
#include "USBDeviceManager.hpp"
#include "AnalogOutputSubsystem.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class AnalogOutputSubsystem implementation

const int AnalogOutputSubsystem::MAX_COUNTS = 65535;
const int AnalogOutputSubsystem::NUM_RANGES = 4;
const double AnalogOutputSubsystem::daRange[][ 2 ] = {
	//  min, range
	  {   0,  5 }								// DAC_RANGE_0_5V
	, {  -5, 10 }								// DAC_RANGE_5V
	, {   0, 10 }								// DAC_RANGE_0_10V
	, { -10, 20 }								// DAC_RANGE_10V
};	// AnalogOutputSubsystem::daRange[]


AnalogOutputSubsystem::AnalogOutputSubsystem( USBDevice &parent )
		: DeviceSubsystem( parent ) {
	range = DAC_RANGE_0_5V;
	numChannels = 0;
	DeviceProperties properties;
	const int result = AIOUSB_GetDeviceProperties( getDeviceIndex(), &properties );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	numChannels = properties.DACChannels;
	assert( numChannels > 0 );
	setRange( DAC_RANGE_0_5V );
}	// AnalogOutputSubsystem::AnalogOutputSubsystem()


AnalogOutputSubsystem::~AnalogOutputSubsystem() {
	// nothing to do
}	// AnalogOutputSubsystem::~AnalogOutputSubsystem()

/*
 * properties
 */

ostream &AnalogOutputSubsystem::print( ostream &out ) {
	assert( &out != 0 );
	out << "    Number of D/A channels: " << dec << numChannels << endl;
	return out;
}	// AnalogOutputSubsystem::print()

/*
 * configuration
 */

AnalogOutputSubsystem &AnalogOutputSubsystem::setRange( int range ) {
	if(
		range < DAC_RANGE_0_5V
		|| range > DAC_RANGE_10V
	)
		throw IllegalArgumentException( "Invalid range" );
	const int result = DACSetBoardRange( getDeviceIndex(), range );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	this->range = range;
	return *this;
}	// AnalogOutputSubsystem::setRange()

/*
 * operations
 */

AnalogOutputSubsystem &AnalogOutputSubsystem::write( int channel, unsigned short counts ) {
	if(
		channel < 0
		|| channel >= numChannels
		|| counts < 0
		|| counts > MAX_COUNTS
	)
		throw IllegalArgumentException( "Invalid channel or counts" );
	const int result = DACDirect( getDeviceIndex(), channel, counts );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogOutputSubsystem::write()


AnalogOutputSubsystem &AnalogOutputSubsystem::write( const UShortArray &points ) {
	if(
		&points == 0
		|| points.size() < 2
		|| ( points.size() & 1 ) != 0
	)
		throw IllegalArgumentException( "Invalid points array" );

	/*
	 * check all channel/count pairs
	 */
	for( int index = 0; index < ( int ) points.size(); index += 2 ) {
		if(
			points.at( index ) < 0
			|| points.at( index ) >= numChannels
			|| points.at( index + 1 ) < 0
			|| points.at( index + 1 ) > MAX_COUNTS
		)
			throw IllegalArgumentException( "Invalid channel or counts" );
	}	// for( int index ...

	const int result = DACMultiDirect( getDeviceIndex(), ( unsigned short * ) points.data(), points.size() / 2 /* channel-count pairs */ );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// AnalogOutputSubsystem::write()


AnalogOutputSubsystem &AnalogOutputSubsystem::writeVolts( int channel, double volts ) {
	return write( channel, voltsToCounts( volts ) );
}	// AnalogOutputSubsystem::writeVolts()


AnalogOutputSubsystem &AnalogOutputSubsystem::writeVolts( const OutputVoltagePointArray &points ) {
	if(
		&points == 0
		|| points.size() < 1
	)
		throw IllegalArgumentException( "Invalid points array" );

	/*
	 * convert channel/voltage pairs to channel/count pairs
	 */
	UShortArray countPoints( points.size() * 2 );
	for( int index = 0; index < ( int ) points.size(); index++ ) {
		if(
			points.at( index ).channel >= 0
			&& points.at( index ).channel < numChannels
		) {
			countPoints.at( index * 2 ) = ( unsigned short ) points.at( index ).channel;
			countPoints.at( index * 2 + 1 ) = voltsToCounts( points.at( index ).volts );
		} else
			throw IllegalArgumentException( "Invalid channel" );
	}	// for( int index ...

	return write( countPoints );
}	// AnalogOutputSubsystem::writeVolts()


double AnalogOutputSubsystem::countsToVolts( unsigned short counts ) const {
	assert( range >= DAC_RANGE_0_5V
		&& range <= DAC_RANGE_10V );
	return ( ( double ) counts / MAX_COUNTS ) * daRange[ range ][ 1 ] + daRange[ range ][ 0 ];
}	// AnalogOutputSubsystem::countsToVolts()


DoubleArray AnalogOutputSubsystem::countsToVolts( const UShortArray &counts ) const {
	if(
		&counts == 0
		|| counts.size() < 1
	)
		throw IllegalArgumentException( "Invalid counts array" );
	DoubleArray volts( counts.size() );
	for( int index = 0; index < ( int ) counts.size(); index++ )
		volts.at( index ) = countsToVolts( counts.at( index ) );
	return volts;
}	// AnalogOutputSubsystem::countsToVolts()


unsigned short AnalogOutputSubsystem::voltsToCounts( double volts ) const {
	assert( range >= DAC_RANGE_0_5V
		&& range <= DAC_RANGE_10V );
	int counts = ( int ) round( MAX_COUNTS * ( volts - daRange[ range ][ 0 ] ) / daRange[ range ][ 1 ] );
	if( counts < 0 )
		counts = 0;
	else if( counts > MAX_COUNTS )
		counts = MAX_COUNTS;
	return ( unsigned short ) counts;
}	// AnalogOutputSubsystem::voltsToCounts()


UShortArray AnalogOutputSubsystem::voltsToCounts( const DoubleArray &volts ) const {
	if(
		&volts == 0
		|| volts.size() < 1
	)
		throw IllegalArgumentException( "Invalid volts array" );
	UShortArray counts( volts.size() );
	for( int index = 0; index < ( int ) volts.size(); index++ )
		counts.at( index ) = voltsToCounts( volts.at( index ) );
	return counts;
}	// AnalogOutputSubsystem::voltsToCounts()

// }}}

}	// namespace AIOUSB


/* end of file */
