/*
 * $RCSfile: CounterSubsystem.cpp,v $
 * $Revision: 1.7 $
 * $Date: 2009/11/23 12:54:12 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class CounterSubsystem implementation
 */


// {{{ includes
#include <assert.h>
#include "../lib/AIOUSB_Core.h"
#include "USBDeviceManager.hpp"
#include "CounterSubsystem.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class Counter implementation

Counter::Counter( CounterSubsystem *parent, int counterIndex ) {
	assert( parent != 0
		&& counterIndex >= 0 );
	this->parent = parent;
	this->counterIndex = counterIndex;
}	// Counter::Counter()


int Counter::getDeviceIndex() const {
	assert( parent != 0 );
	return parent->getDeviceIndex();
}	// Counter::getDeviceIndex()


Counter &Counter::setMode( int mode ) {
	assert( counterIndex >= 0 );
	if(
		mode < MODE_TERMINAL_COUNT
		|| mode > MODE_HW_TRIGGERED
	)
		throw IllegalArgumentException( "Invalid mode" );
	const int result = CTR_8254Mode( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, mode );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// Counter::setMode()


Counter &Counter::setCount( unsigned short count ) {
	assert( counterIndex >= 0 );
	const int result = CTR_8254Load( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, count );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// Counter::setCount()


Counter &Counter::setModeAndCount( int mode, unsigned short count ) {
	assert( counterIndex >= 0 );
	if(
		mode < MODE_TERMINAL_COUNT
		|| mode > MODE_HW_TRIGGERED
	)
		throw IllegalArgumentException( "Invalid mode" );
	const int result = CTR_8254ModeLoad( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, mode, count );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// Counter::setModeAndCount()


unsigned short Counter::readCount() {
	assert( counterIndex >= 0 );
	unsigned short count;
	const int result = CTR_8254Read( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, &count );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return count;
}	// Counter::ReadCount()


UShortArray Counter::readCountAndStatus() {
	assert( counterIndex >= 0 );
	unsigned short count;
	unsigned char status;
	const int result = CTR_8254ReadStatus( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, &count, &status );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	UShortArray values( 2 );
	values.at( 0 ) = count;
	values.at( 1 ) = status;
	return values;
}	// Counter::readCountAndStatus()


unsigned short Counter::readCountAndSetModeAndCount( int mode, unsigned short count ) {
	assert( counterIndex >= 0 );
	if(
		mode < MODE_TERMINAL_COUNT
		|| mode > MODE_HW_TRIGGERED
	)
		throw IllegalArgumentException( "Invalid mode" );
	unsigned short prevCount;
	const int result = CTR_8254ReadModeLoad( getDeviceIndex(), 0 /* BlockIndex */, counterIndex, mode, count, &prevCount );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return prevCount;
}	// Counter::readCountAndSetModeAndCount()

// }}}

// {{{ class CounterSubsystem implementation

CounterSubsystem::CounterSubsystem( USBDevice &parent )
		: DeviceSubsystem( parent ) {
	numCounterBlocks = numCounters = 0;
	DeviceProperties properties;
	const int result = AIOUSB_GetDeviceProperties( getDeviceIndex(), &properties );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	numCounterBlocks = properties.Counters;
	assert( numCounterBlocks > 0 );
	if( numCounterBlocks > 20 )
		numCounterBlocks = 20;					// limit to arbitrary, reasonable number
	numCounters = numCounterBlocks * COUNTERS_PER_BLOCK;
	for( int index = 0; index < numCounters; index++ )
		counters.insert( counters.end(), new Counter( this, index ) );
	assert( ( int ) counters.size() == numCounters );
}	// CounterSubsystem::CounterSubsystem()


CounterSubsystem::~CounterSubsystem() {
	assert( ( int ) counters.size() == numCounters );
	for( int index = 0; index < numCounters; index++ ) {
		assert( counters.at( index ) != 0 );
		delete counters.at( index );
	}	// for( int index ...
	counters.clear();
}	// CounterSubsystem::~CounterSubsystem()

/*
 * properties
 */

ostream &CounterSubsystem::print( ostream &out ) {
	assert( &out != 0 );
	out
		<< "    Number of counter blocks: " << numCounterBlocks << endl
		<< "    Number of counters: " << numCounters << endl;
	return out;
}	// CounterSubsystem::print()

/*
 * operations
 */

Counter &CounterSubsystem::getCounter( int counter ) {
	if(
		counter < 0
		|| counter >= numCounters
	)
		throw IllegalArgumentException( "Invalid counter" );
	return *counters.at( counter );
}	// CounterSubsystem::getCounter()


UShortArray CounterSubsystem::readCounts( bool oldData ) {
	UShortArray counts( oldData ? numCounters : ( numCounters + 1 ) );
	const int result = oldData
		? CTR_8254ReadLatched( getDeviceIndex(), counts.data() )	// last word of 'counts' is non-zero if data is "old"
		: CTR_8254ReadAll( getDeviceIndex(), counts.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return counts;
}	// CounterSubsystem::readCounts()


CounterSubsystem &CounterSubsystem::selectGate( int counter ) {
	if(
		counter < 0
		|| counter >= numCounters
	)
		throw IllegalArgumentException( "Invalid counter" );
	const int result = CTR_8254SelectGate( getDeviceIndex(), counter );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// CounterSubsystem::selectGate()


double CounterSubsystem::startClock( int counterBlock, double clockHz ) {
	if(
		counterBlock < 0
		|| counterBlock >= numCounterBlocks
		|| clockHz < 0
	)
		throw IllegalArgumentException( "Invalid counter block or frequency" );
	const int result = CTR_StartOutputFreq( getDeviceIndex(), counterBlock, &clockHz );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return clockHz;
}	// CounterSubsystem::startClock()

// }}}

}	// namespace AIOUSB


/* end of file */
