/*
 * $RCSfile: DigitalIOSubsystem.cpp,v $
 * $Revision: 1.9 $
 * $Date: 2009/11/30 20:04:51 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class DigitalIOSubsystem implementation
 */


// {{{ includes
#include <assert.h>
#include "../lib/AIOUSB_Core.h"
#include "USBDeviceManager.hpp"
#include "DigitalIOSubsystem.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class DigitalIOSubsystem implementation

DigitalIOSubsystem::DigitalIOSubsystem( USBDevice &parent )
		: DeviceSubsystem( parent ) {
	numPorts
		= numChannels
		= numTristateGroups
		= numTristates
		= 0;
	DeviceProperties properties;
	const int result = AIOUSB_GetDeviceProperties( getDeviceIndex(), &properties );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	numPorts = properties.DIOPorts;
	numTristateGroups = properties.Tristates;
	assert( numPorts > 0 );
	numChannels = numPorts * BITS_PER_BYTE;
	numTristates = numTristateGroups * BITS_PER_BYTE;
	writeValues.resize( numPorts );
}	// DigitalIOSubsystem::DigitalIOSubsystem()


DigitalIOSubsystem::~DigitalIOSubsystem() {
	// nothing to do
}	// DigitalIOSubsystem::~DigitalIOSubsystem()

/*
 * properties
 */

ostream &DigitalIOSubsystem::print( ostream &out ) {
	assert( &out != 0 );
	out
		<< "    Number of digital I/O ports: " << dec << numPorts << endl
		<< "    Number of digital I/O channels: " << numChannels << endl
		<< "    Number of tristate groups: " << numTristateGroups << endl
		<< "    Number of tristates: " << numTristates << endl;
	return out;
}	// DigitalIOSubsystem::print()

/*
 * configuration
 */

DigitalIOSubsystem &DigitalIOSubsystem::configure( bool tristate, const BoolArray &outputs, const BoolArray &values ) {
	if(
		&outputs == 0
		|| outputs.size() < 1
		|| ( int ) outputs.size() > numPorts
		|| &values == 0
		|| values.size() < 1
		|| ( int ) values.size() > numChannels
	)
		throw IllegalArgumentException( "Invalid outputs or values" );
	UCharArray outputMask( ( numPorts + BITS_PER_BYTE - 1 ) / BITS_PER_BYTE );
	const int result = DIO_Configure( getDeviceIndex()
		, ( unsigned char ) tristate
		, bitsToBytes( outputMask, 0, outputs ).data()
		, bitsToBytes( writeValues, 0, values ).data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// DigitalIOSubsystem::configure()


DigitalIOSubsystem &DigitalIOSubsystem::configure( const BoolArray &tristates, const BoolArray &outputs, const BoolArray &values ) {
	if(
		&tristates == 0
		|| tristates.size() < 1
		|| ( int ) tristates.size() > numTristates
		|| &outputs == 0
		|| outputs.size() < 1
		|| ( int ) outputs.size() > numPorts
		|| &values == 0
		|| values.size() < 1
		|| ( int ) values.size() > numChannels
	)
		throw IllegalArgumentException( "Invalid tristates, outputs or values" );
	UCharArray outputMask( ( numPorts + BITS_PER_BYTE - 1 ) / BITS_PER_BYTE );
	UCharArray tristateMask( numTristateGroups );
	const int result = DIO_ConfigureEx( getDeviceIndex()
		, bitsToBytes( outputMask, 0, outputs ).data()
		, bitsToBytes( writeValues, 0, values ).data()
		, bitsToBytes( tristateMask, 0, tristates ).data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// DigitalIOSubsystem::configure()


DigitalIOSubsystem &DigitalIOSubsystem::getConfiguration( BoolArray &tristates, BoolArray &outputs ) {
	if(
		&tristates == 0
		|| tristates.size() < 1
		|| ( int ) tristates.size() > numTristates
		|| &outputs == 0
		|| outputs.size() < 1
		|| ( int ) outputs.size() > numPorts
	)
		throw IllegalArgumentException( "Invalid tristates or outputs" );
	UCharArray outputMask( ( numPorts + BITS_PER_BYTE - 1 ) / BITS_PER_BYTE );
	UCharArray tristateMask( numTristateGroups );
	const int result = DIO_ConfigurationQuery( getDeviceIndex(), outputMask.data(), tristateMask.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	bytesToBits( tristates, tristateMask, 0 );
	bytesToBits( outputs, outputMask, 0 );
	return *this;
}	// DigitalIOSubsystem::getConfiguration()

/*
 * operations
 */

bool DigitalIOSubsystem::read( int channel ) {
	if(
		channel < 0
		|| channel >= numChannels
	)
		throw IllegalArgumentException( "Invalid channel" );
	unsigned char byteValue;
	const int result = DIO_Read1( getDeviceIndex(), channel, &byteValue );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return byteValue != 0;
}	// DigitalIOSubsystem::read()


BoolArray DigitalIOSubsystem::read( int startChannel, int numChannels ) {
	if(
		numChannels < 1
		|| startChannel < 0
		|| startChannel + numChannels > this->numChannels
	)
		throw IllegalArgumentException( "Invalid start channel or number of channels" );
	UCharArray ports( numPorts );
	const int result = DIO_ReadAll( getDeviceIndex(), ports.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	BoolArray values( numChannels );
	return bytesToBits( values, ports, startChannel );
}	// DigitalIOSubsystem::read()


DigitalIOSubsystem &DigitalIOSubsystem::write( int channel, bool value ) {
	BoolArray values( 1 );
	values.at( 0 ) = value;
	return write( channel, values );
}	// DigitalIOSubsystem::write()


DigitalIOSubsystem &DigitalIOSubsystem::write( int startChannel, const BoolArray &values ) {
	if(
		&values == 0
		|| values.size() < 1
		|| startChannel < 0
		|| startChannel + ( int ) values.size() > numChannels
	)
		throw IllegalArgumentException( "Invalid values or start channel" );
	const int result = DIO_WriteAll( getDeviceIndex(), bitsToBytes( writeValues, startChannel, values ).data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// DigitalIOSubsystem::write()

/*
 * utilities
 */

UCharArray &DigitalIOSubsystem::bitsToBytes( UCharArray &dest, int bit, const BoolArray &src ) {
	assert( &dest != 0
		&& dest.size() > 0
		&& &src != 0
		&& src.size() > 0
		&& bit >= 0
		&& ( bit + src.size() ) <= dest.size() * BITS_PER_BYTE );
	int destByte = bit / BITS_PER_BYTE;
	unsigned char destBit = ( unsigned char ) ( 1 << ( bit % BITS_PER_BYTE ) );
	const unsigned char maxDestBit = ( unsigned char ) ( 1 << ( BITS_PER_BYTE - 1 ) );
	for( int index = 0; index < ( int ) src.size(); index++ ) {
		if( src.at( index ) )
			dest.at( destByte ) |= destBit;
		else
			dest.at( destByte ) &= ~destBit;
		if( ( destBit & maxDestBit ) == 0 ) {
			destBit <<= 1;
		} else {
			// advance to next byte
			destBit = 1;
			destByte++;
		}	// if( ( destBit ...
	}	// for( int index ...
	return dest;
}	// DigitalIOSubsystem::bitsToBytes()


BoolArray &DigitalIOSubsystem::bytesToBits( BoolArray &dest, const UCharArray &src, int bit ) {
	assert( &dest != 0
		&& dest.size() > 0
		&& &src != 0
		&& src.size() > 0
		&& bit >= 0
		&& ( bit + dest.size() ) <= src.size() * BITS_PER_BYTE );
	int srcByte = bit / BITS_PER_BYTE;
	unsigned char srcBit = ( unsigned char ) ( 1 << ( bit % BITS_PER_BYTE ) );
	const unsigned char maxSrcBit = ( unsigned char ) ( 1 << ( BITS_PER_BYTE - 1 ) );
	for( int index = 0; index < ( int ) dest.size(); index++ ) {
		dest.at( index ) = ( src.at( srcByte ) & srcBit ) != 0;
		if( ( srcBit & maxSrcBit ) == 0 ) {
			srcBit <<= 1;
		} else {
			// advance to next byte
			srcBit = 1;
			srcByte++;
		}	// if( ( srcBit ...
	}	// for( int index ...
	return dest;
}	// DigitalIOSubsystem::bytesToBits()

// }}}

}	// namespace AIOUSB


/* end of file */
