/*
 * $RCSfile: USBDevice.cpp,v $
 * $Revision: 1.10 $
 * $Date: 2009/11/22 20:37:37 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class USBDevice implementation
 */


// {{{ includes
#include <iomanip>
#include <assert.h>
#include "../lib/AIOUSB_Core.h"
#include "USBDevice.hpp"
#include "USBDeviceManager.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class USBDevice implementation

const int USBDevice::CUSTOM_EEPROM_SIZE = 0x200;


USBDevice::USBDevice( int productID, int deviceIndex ) {
	if(
		productID < 0
		|| productID > 0xffff
		|| deviceIndex < 0
		|| deviceIndex > 100					// arbitrary sanity check
	)
		throw IllegalArgumentException( "Invalid product ID" );
	this->productID = productID;
	this->deviceIndex = deviceIndex;
	serialNumber = 0;
	DeviceProperties properties;
	const int result = AIOUSB_GetDeviceProperties( deviceIndex, &properties );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	name = properties.Name;
	serialNumber = properties.SerialNumber;
}	// USBDevice::USBDevice()


USBDevice::~USBDevice() {
	// nothing to do
}	// USBDevice::~USBDevice()

/*
 * properties
 */

ostream &USBDevice::print( ostream &out ) {
	assert( &out != 0 );
	out
		<< "  Device at index " << left << dec << deviceIndex << ':' << endl
		<< "    Product ID: 0x" << right << setw( 4 ) << setfill( '0' ) << hex << productID << endl
		<< "    Product name: " << left << getName() << endl
		<< "    Serial number: 0x" << right << setw( 16 ) << hex << getSerialNumber() << endl;
	return out;
}	// USBDevice::print()

/*
 * configuration
 */

USBDevice &USBDevice::clearFIFO( int method ) {
	if(
		method != CLEAR_FIFO_METHOD_IMMEDIATE
		&& method != CLEAR_FIFO_METHOD_AUTO
		&& method != CLEAR_FIFO_METHOD_IMMEDIATE_AND_ABORT
		&& method != CLEAR_FIFO_METHOD_WAIT
	)
		throw IllegalArgumentException( "Invalid method" );
	const int result = AIOUSB_ClearFIFO( deviceIndex, method );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// USBDevice::clearFIFO()


USBDevice &USBDevice::setMiscClock( double clockHz ) {
	if( clockHz < 0 )
		throw IllegalArgumentException( "Invalid clock frequency" );
	const int result = AIOUSB_SetMiscClock( deviceIndex, clockHz );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// USBDevice::setMiscClock()


int USBDevice::getCommTimeout() const {
	return AIOUSB_GetCommTimeout( deviceIndex );
}	// USBDevice::getCommTimeout()


USBDevice &USBDevice::setCommTimeout( int timeout ) {
	if(
		timeout < 0
		|| timeout > 100000						// arbitrary sanity check
	)
		throw IllegalArgumentException( "Invalid timeout" );
	const int result = AIOUSB_SetCommTimeout( deviceIndex, timeout );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// USBDevice::setCommTimeout()

/*
 * operations
 */

USBDevice &USBDevice::reset() {
	const int result = ( ResultCode ) AIOUSB_Reset( deviceIndex );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// USBDevice::reset()


USBDevice &USBDevice::customEEPROMWrite( int address, const UCharArray &data ) {
	if(
		&data == 0
		|| data.size() < 1
		|| address < 0
		|| address + ( int ) data.size() > CUSTOM_EEPROM_SIZE
	)
		throw IllegalArgumentException( "Invalid data or address" );
	const int result = AIOUSB::CustomEEPROMWrite( deviceIndex, address, data.size(), ( void * ) data.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return *this;
}	// USBDevice::customEEPROMWrite()


UCharArray USBDevice::customEEPROMRead( int address, int numBytes ) {
	if(
		numBytes < 1
		|| address < 0
		|| address + numBytes > CUSTOM_EEPROM_SIZE
	)
		throw IllegalArgumentException( "Invalid data or address" );
	UCharArray data( numBytes );
	unsigned long size = data.size();
	const int result = AIOUSB::CustomEEPROMRead( deviceIndex, address, &size, data.data() );
	if( result != AIOUSB_SUCCESS )
		throw OperationFailedException( result );
	return data;
}	// USBDevice::customEEPROMRead()

// }}}

// {{{ function implementation

ostream &operator<<( ostream &out, USBDevice &device ) {
	assert( &device != 0 );
	return device.print( out );
}	// operator<<()


ostream &operator<<( ostream &out, USBDevice *device ) {
	assert( device != 0 );
	if( device != 0 )
		device->print( out );
	return out;
}	// operator<<()

// }}}

}	// namespace AIOUSB


/* end of file */
