/*
 * $RCSfile: USBDeviceManager.cpp,v $
 * $Revision: 1.17 $
 * $Date: 2009/11/30 20:07:54 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * class USBDeviceManager implementation
 */


// {{{ includes
#include <iostream>
#include <iterator>
#include <assert.h>
#include "USBDeviceManager.hpp"
#include "USB_AI16_Family.hpp"
#include "USB_AO16_Family.hpp"
#include "USB_CTR_15_Family.hpp"
#include "USB_DIO_16_Family.hpp"
#include "USB_DIO_32_Family.hpp"
#include "USB_DIO_Family.hpp"
// }}}

using namespace std;

namespace AIOUSB {

// {{{ class USBDeviceManager implementation

USBDeviceManager::USBDeviceManager()
		: VERSION_NUMBER( "1.4" )
		, VERSION_DATE( "30 November 2009" )
		, MESSAGE_NOT_OPEN( "Not open, must call open() first" )
		, OPEN_PATTERN( 0x786938f5 /* unique, random value */ ) {
	openStatus = 0;
}	// USBDeviceManager::USBDeviceManager()


USBDeviceManager::~USBDeviceManager() {
	if( isOpen() )
		close();								// in case user forgot
}	// USBDeviceManager::~USBDeviceManager()

/*
 * properties
 */

ostream &USBDeviceManager::print( ostream &out ) {
	assert( &out != 0 );
	if( ! isOpen() )
		throw OperationFailedException( MESSAGE_NOT_OPEN );
	if( deviceList.size() > 0 ) {
		out << "ACCES devices found:" << endl;
		copy( deviceList.begin(), deviceList.end(), ostream_iterator<USBDevice *>( out ) );
	} else
		out << "No ACCES devices found" << endl;
	return out;
}	// USBDeviceManager::print()


USBDeviceManager &USBDeviceManager::printDevices() {
	print( cout );
	return *this;
}	// USBDeviceManager::printDevices()

/*
 * operations
 */

USBDeviceManager &USBDeviceManager::open() {
	if( ! isOpen() ) {
		const int result = AIOUSB_Init();
		if( result == AIOUSB_SUCCESS ) {
			openStatus = OPEN_PATTERN;
			scanForDevices();
		} else
			throw OperationFailedException( result );
	} else
		throw OperationFailedException( "Already open" );
	return *this;
}	// USBDeviceManager::open()


USBDeviceManager &USBDeviceManager::close() {
	if( isOpen() ) {
		emptyDeviceList();
		AIOUSB_Exit();
		openStatus = 0;
	} else
		throw OperationFailedException( MESSAGE_NOT_OPEN );
	return *this;
}	// USBDeviceManager::close()


void USBDeviceManager::emptyDeviceList() {
	for( int index = 0; index < ( int ) deviceList.size(); index++ ) {
		assert( deviceList.at( index ) != 0 );
		delete deviceList.at( index );
	}	// for( int index ...
	deviceList.clear();
}	// USBDeviceManager::emptyDeviceList()


USBDeviceManager &USBDeviceManager::scanForDevices() {
	if( isOpen() ) {
		emptyDeviceList();
		const int MAX_DEVICES = 100;
		int devices[ 1 + MAX_DEVICES * 2 ];		// device index-product ID pairs
		const int result = AIOUSB_GetDeviceByProductID( 0, 0xffff, MAX_DEVICES, devices );		// get all devices
		if( result != AIOUSB_SUCCESS )
			throw OperationFailedException( result );
		const int numDevices = devices[ 0 ];
		for( int index = 0; index < numDevices; index++ ) {
			USBDevice *device = 0;
			const int deviceIndex = devices[ 1 + index * 2 ];
			const int productID = devices[ 1 + index * 2 + 1 ];
			if( productID == USB_CTR_15 ) {
				device = new USB_CTR_15_Family( productID, deviceIndex );
			} else if( productID == USB_DIO_32 ) {
				device = new USB_DIO_32_Family( productID, deviceIndex );
			} else if(
				productID == USB_DIO_48
				|| productID == USB_DIO_96
				|| productID == USB_IIRO_16
				|| productID == USB_II_16
				|| productID == USB_RO_16
				|| productID == USB_IIRO_8
				|| productID == USB_II_8
				|| productID == USB_IIRO_4
				|| productID == USB_IDIO_16
				|| productID == USB_II_16_OLD
				|| productID == USB_IDO_16
				|| productID == USB_IDIO_8
				|| productID == USB_II_8_OLD
				|| productID == USB_IDIO_4
				|| productID == USB_IIRO4_2SM
				|| productID == USB_IIRO4_COM
				|| productID == USB_DIO16RO8
				|| productID == PICO_DIO16RO8
			) {
				device = new USB_DIO_Family( productID, deviceIndex );
			} else if(
				(
					productID >= USB_DI16A_REV_A1
					&& productID <= USB_DI16A_REV_A2
				)
				|| (
					productID >= USB_DIO_16H
					&& productID <= USB_DIO_16A
				)
			) {
				device = new USB_DIO_16_Family( productID, deviceIndex );
			} else if(
				productID >= USB_AI16_16A
				&& productID <= USB_AI12_128E
			) {
				device = new USB_AI16_Family( productID, deviceIndex );
			} else if(
				productID >= USB_AO16_16A
				&& productID <= USB_AO12_4
			) {
				device = new USB_AO16_Family( productID, deviceIndex );
			}	// if( productID ...
			if( device != 0 )
				deviceList.insert( deviceList.end(), device );
		}	// for( int index ...
	} else
		throw OperationFailedException( MESSAGE_NOT_OPEN );
	return *this;
}	// USBDeviceManager::scanForDevices()


DeviceList USBDeviceManager::getDeviceByProductID( int productID ) const {
	return getDeviceByProductID( productID, productID );
}	// USBDeviceManager::getDeviceByProductID()


DeviceList USBDeviceManager::getDeviceByProductID( int minProductID, int maxProductID ) const {
	if(
		minProductID < 0
		|| minProductID > 0xffff
		|| maxProductID < minProductID
		|| maxProductID > 0xffff
	)
		throw IllegalArgumentException( "Invalid product IDs" );
	DeviceList devices;
	for( int index = 0; index < ( int ) deviceList.size(); index++ ) {
		assert( deviceList.at( index ) != 0 );
		const int productID = deviceList.at( index )->getProductID();
		if(
			productID >= minProductID
			&& productID <= maxProductID
		)
			devices.insert( devices.end(), deviceList.at( index ) );
	}	// for( int index ...
	return devices;
}	// USBDeviceManager::getDeviceByProductID()


/*
 * in theory, there should be at most one device with the specified serial number, but just in case ...
 */
DeviceList USBDeviceManager::getDeviceBySerialNumber( __uint64_t serialNumber ) const {
	DeviceList devices;
	for( int index = 0; index < ( int ) deviceList.size(); index++ ) {
		assert( deviceList.at( index ) != 0 );
		if( deviceList.at( index )->getSerialNumber() == serialNumber )
			devices.insert( devices.end(), deviceList.at( index ) );
	}	// for( int index ...
	return devices;
}	// USBDeviceManager::getDeviceBySerialNumber()

// }}}

}	// namespace AIOUSB


/* end of file */
