/*
 * $RCSfile: AnalogInputSubsystem.java,v $
 * $Date: 2009/11/29 17:11:14 $
 * $Revision: 1.14 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 */

package com.acces.aiousb;

// {{{ imports
import java.io.*;
import java.util.*;
// }}}

/**
 * Class AnalogInputSubsystem represents the analog input subsystem of a device. One accesses
 * this analog input subsystem through its parent object, typically through a method such as
 * <i>adc() (see {@link USB_AI16_Family#adc()})</i>.
 */

public class AnalogInputSubsystem extends DeviceSubsystem {

	// {{{ public constants
	/*
	 * A/D calibration modes; if ground or reference mode is selected, only one A/D
	 * sample may be taken at a time; that means, one channel and no oversampling;
	 * attempting to read more than one channel or use an oversample setting of more
	 * than zero will result in a timeout error
	 */
	/** Selects normal measurement mode <i>(see {@link #setCalMode( int mode ) setCalMode()})</i>. */
	public static final int CAL_MODE_NORMAL			= 0;		// normal measurement
	/** Selects ground calibration mode. */
	public static final int CAL_MODE_GROUND			= 1;		// measure ground
	/** Selects reference (full scale) calibration mode. */
	public static final int CAL_MODE_REFERENCE		= 3;		// measure reference

	/*
	 * A/D trigger and counter bits
	 */
	/** If set, counter 0 is externally triggered <i>(see {@link #setTriggerMode( int mode ) setTriggerMode()})</i>. */
	public static final int TRIG_MODE_CTR0_EXT		= 0x10;		// 1 == counter 0 externally triggered
	/** If set, the A/D is triggered by the falling edge of its trigger source, otherwise it's triggered by the rising edge. */
	public static final int TRIG_MODE_FALLING_EDGE	= 0x08;		// 1 == triggered by falling edge
	/** If set, each trigger will cause the A/D to scan all the channels, otherwise the A/D will read a single channel with each trigger. */
	public static final int TRIG_MODE_SCAN			= 0x04;		// 1 == scan all channels, 0==single channel
	/** If set, the A/D is triggered by an external pin on the board. */
	public static final int TRIG_MODE_EXTERNAL		= 0x02;		// 1 == triggered by external pin on board
	/** If set, the A/D is triggered by counter 2. */
	public static final int TRIG_MODE_TIMER			= 0x01;		// 1 == triggered by counter 2

	/** Unipolar, 0-10 volt range <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. */
	public static final int GAIN_CODE_0_10V			= 0;		// 0-10V
	/** Bipolar, -10 to +10 volt range. */
	public static final int GAIN_CODE_10V			= 1;		// +/-10V
	/** Unipolar, 0-5 volt range. */
	public static final int GAIN_CODE_0_5V			= 2;		// 0-5V
	/** Bipolar, -5 to +5 volt range. */
	public static final int GAIN_CODE_5V			= 3;		// +/-5V
	/** Unipolar, 0-2 volt range. */
	public static final int GAIN_CODE_0_2V			= 4;		// 0-2V
	/** Bipolar, -2 to +2 volt range. */
	public static final int GAIN_CODE_2V			= 5;		// +/-2V
	/** Unipolar, 0-1 volt range. */
	public static final int GAIN_CODE_0_1V			= 6;		// 0-1V
	/** Bipolar, -1 to +1 volt range. */
	public static final int GAIN_CODE_1V			= 7;		// +/-1V

	/** Maximum number of counts A/D can read. */
	public static final int MAX_COUNTS				= 65535;
	// }}}

	// {{{ protected members
	/*
	 * register indexes
	 */
	protected static final int NUM_CONFIG_REGISTERS	= 20;		// number of "registers" (bytes) in config. block
	protected static final int NUM_MUX_CONFIG_REGISTERS = 21;	// number of "registers" (bytes) in config. block when MUX installed
	protected static final int NUM_GAIN_CODE_REGISTERS = 16;	// number of gain code registers in config. block
	protected static final int REG_GAIN_CODE		= 0;		// gain code for A/D channel 0 (one of GAIN_CODE_* settings below)
																// gain codes for channels 1-15 occupy configuration block indexes 1-15
	protected static final int REG_CAL_MODE			= 16;		// calibration mode (one of CAL_MODE_* settings below)
	protected static final int REG_TRIG_MODE		= 17;		// trigger mode (one of TRIG_MODE_* settings below)
	protected static final int REG_START_END		= 18;		// start and end channels for scan (bits 7-4 contain end channel, bits 3-0 contain start channel)
	protected static final int REG_OVERSAMPLE		= 19;		// oversample setting (0-255 samples in addition to single sample)
	protected static final int REG_MUX_START_END	= 20;		// MUX start and end channels for scan (bits 7-4 contain end channel MS-nibble, bits 3-0 contain start channel MS-nibble)

	protected static final int NUM_GAIN_CODES		= 8;
	protected static final int GAIN_CODE_MIN		= 0;
	protected static final int GAIN_CODE_MAX		= 7;
	protected static final int GAIN_CODE_VALID_MASK	= 7;
	protected static final int DIFFERENTIAL_MODE	= 8;		// OR with gain code to enable differential mode for channels 0-7
	protected static final int MAX_OVERSAMPLE		= 0xff;
	protected static final int TRIG_MODE_VALID_MASK	= (
		TRIG_MODE_CTR0_EXT
		| TRIG_MODE_FALLING_EDGE
		| TRIG_MODE_SCAN
		| TRIG_MODE_EXTERNAL
		| TRIG_MODE_TIMER );
	protected static final int CAL_TABLE_WORDS		= 64 * 1024;	// 64K 2-byte words

	protected static final int AUTO_CAL_UNKNOWN		= 0;
	protected static final int AUTO_CAL_NOT_PRESENT	= 1;
	protected static final int AUTO_CAL_PRESENT		= 2;

	protected int numChannels;									// number of A/D channels
	protected int numMUXChannels;								// number of MUXed A/D channels
	protected int autoCalFeature = AUTO_CAL_UNKNOWN;			// cached A/D auto. cal. feature
	protected byte[] configBlock;								// registers that comprise configuration block
	protected byte[] prevConfigBlock;							// previous configuration block (so we can detect changes)
	protected int readBulkSamplesRequested;						// total number of samples requested during bulk read procedure
	protected int readBulkSamplesRetrieved;						// number of samples retrieved so far during  bulk read procedure
	// }}}

	// {{{ protected methods

	protected native int getConfigSize( int deviceIndex );
	protected native int getNumChannels( int deviceIndex );
	protected native int getNumMUXChannels( int deviceIndex );
	protected native int queryCal( int deviceIndex );
		// => unsigned long ADC_QueryCal( unsigned long DeviceIndex )
	protected native boolean isDiscardFirstSample( int deviceIndex );
		// => AIOUSB_BOOL AIOUSB_IsDiscardFirstSample( unsigned long DeviceIndex )
	protected native int setDiscardFirstSample( int deviceIndex, boolean discard );
		// => unsigned long AIOUSB_SetDiscardFirstSample( unsigned long DeviceIndex, AIOUSB_BOOL discard )
	protected native int loadCalTable( int deviceIndex, String fileName );
		// => unsigned long AIOUSB_ADC_LoadCalTable( unsigned long DeviceIndex, const char *fileName )
	protected native int setCalTable( int deviceIndex, char[] calTable );
		// => unsigned long AIOUSB_ADC_SetCalTable( unsigned long DeviceIndex, const unsigned short calTable[] )
	protected native int setStreamingBlockSize( int deviceIndex, int blockSize );
		// => unsigned long AIOUSB_SetStreamingBlockSize( unsigned long DeviceIndex, unsigned long BlockSize )
	protected native int getConfig( int deviceIndex, byte[] configBlock );
		// => unsigned long ADC_GetConfig( unsigned long DeviceIndex, unsigned char *pConfigBuf, unsigned long *ConfigBufSize )
	protected native int setConfig( int deviceIndex, byte[] configBlock );
		// => unsigned long ADC_SetConfig( unsigned long DeviceIndex, unsigned char *pConfigBuf, unsigned long *ConfigBufSize )
	protected native int internalCal( int deviceIndex, boolean autoCal, char[] returnCalTable, String saveFileName );
		// => unsigned long AIOUSB_ADC_InternalCal( unsigned long DeviceIndex, AIOUSB_BOOL autoCal, unsigned short returnCalTable[], const char *saveFileName )
	protected native int externalCal( int deviceIndex, double[] points, char[] returnCalTable, String saveFileName );
		// => unsigned long AIOUSB_ADC_ExternalCal( unsigned long DeviceIndex, const double points[], int numPoints, unsigned short returnCalTable[], const char *saveFileName )
	protected native int getScan( int deviceIndex, char[] counts );
		// => unsigned long AIOUSB_GetScan( unsigned long DeviceIndex, unsigned short counts[] )
	protected native int readBulkStart( int deviceIndex, int numSamples );
	protected native int readBulkSamplesRemaining( int deviceIndex, int[] samplesRemaining );
	protected native int readBulkNext( int deviceIndex, int offset, char[] counts );
	protected native int readBulkFinish( int deviceIndex );
	protected native double countsToVolts( int deviceIndex, int channel, char counts );
		// => double AIOUSB_CountsToVolts( unsigned long DeviceIndex, unsigned channel, unsigned short counts )
	protected native int countsToVolts( int deviceIndex, int startChannel, char[] counts, double[] volts );
		// => unsigned long AIOUSB_ArrayCountsToVolts( unsigned long DeviceIndex, int startChannel, int numChannels, const unsigned short counts[], double volts[] )
	protected native char voltsToCounts( int deviceIndex, int channel, double volts );
		// => unsigned short AIOUSB_VoltsToCounts( unsigned long DeviceIndex, unsigned channel, double volts )
	protected native int voltsToCounts( int deviceIndex, int startChannel, double[] volts, char[] counts );
		// => unsigned long AIOUSB_ArrayVoltsToCounts( unsigned long DeviceIndex, int startChannel, int numChannels, const double volts[], unsigned short counts[] )

	protected void validateConfigBlock() {
		assert configBlock != null
			&& (
				configBlock.length == NUM_CONFIG_REGISTERS
				|| configBlock.length == NUM_MUX_CONFIG_REGISTERS
			);
	}	// validateConfigBlock()

	protected AnalogInputSubsystem writeConfig( boolean force ) {
		validateConfigBlock();
		assert prevConfigBlock != null
			&& (
				prevConfigBlock.length == NUM_CONFIG_REGISTERS
				|| prevConfigBlock.length == NUM_MUX_CONFIG_REGISTERS
			);
		if(
			force
			|| ! Arrays.equals( prevConfigBlock, configBlock )
		) {
			/*
			 * configuration block has been modified, send to device
			 */
			System.arraycopy( configBlock, 0, prevConfigBlock, 0, configBlock.length );
			final int result = setConfig( getDeviceIndex(), prevConfigBlock );
			if( result != USBDeviceManager.SUCCESS )
				throw new OperationFailedException( result );
		}	// if( force ...
		return this;
	}	// writeConfig()

	protected int getStartChannel() {
		validateConfigBlock();
		int startChannel;
		if( configBlock.length == NUM_MUX_CONFIG_REGISTERS )
			startChannel
				= ( ( configBlock[ REG_MUX_START_END ] & 0x0f ) << 4 )
				| ( configBlock[ REG_START_END ] & 0xf );
		else
			startChannel = configBlock[ REG_START_END ] & 0xf;
		return startChannel;
	}	// getStartChannel()

	protected int getEndChannel() {
		validateConfigBlock();
		int endChannel;
		if( configBlock.length == NUM_MUX_CONFIG_REGISTERS )
			endChannel
				= ( configBlock[ REG_MUX_START_END ] & 0xf0 )
				| ( ( configBlock[ REG_START_END ] & 0xff ) >>> 4 );
		else
			endChannel = ( configBlock[ REG_START_END ] & 0xff ) >>> 4;
		return endChannel;
	}	// getEndChannel()

	protected AnalogInputSubsystem setScanRange( int startChannel, int numChannels ) {
		validateConfigBlock();
		if( configBlock.length == NUM_MUX_CONFIG_REGISTERS ) {
			/*
			 * this board has a MUX, so support more channels
			 */
			if(
				numChannels < 1
				|| startChannel < 0
				|| startChannel + numChannels > numMUXChannels
			)
				throw new IllegalArgumentException( "Invalid start channel: " + startChannel
					+ ", or number of channels: " + numChannels );
			final int endChannel = startChannel + numChannels - 1;
			configBlock[ REG_START_END ] = ( byte ) ( ( endChannel << 4 ) | ( startChannel & 0x0f ) );
			configBlock[ REG_MUX_START_END ] = ( byte ) ( ( endChannel & 0xf0 ) | ( ( startChannel >>> 4 ) & 0x0f ) );
		} else {
			/*
			 * this board doesn't have a MUX, so support base number of channels
			 */
			if(
				numChannels < 1
				|| startChannel < 0
				|| startChannel + numChannels > this.numChannels
			)
				throw new IllegalArgumentException( "Invalid start channel: " + startChannel
					+ ", or number of channels: " + numChannels );
			final int endChannel = startChannel + numChannels - 1;
			configBlock[ REG_START_END ] = ( byte ) ( ( endChannel << 4 ) | startChannel );
		}	// if( configBlock.length ...
		return writeConfig( false );
	}	// setScanRange()

	// }}}

	// {{{ public methods

	/**
	 * Constructor for analog input subsystem.
	 * @param parent the parent USB device object that this subsystem is a part of.
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem( USBDevice parent ) {
		super( parent );
		numChannels = getNumChannels( getDeviceIndex() );
		assert numChannels > 0;
		numMUXChannels = getNumMUXChannels( getDeviceIndex() );
		configBlock = new byte[ getConfigSize( getDeviceIndex() ) ];
		validateConfigBlock();
		prevConfigBlock = new byte[ configBlock.length ];
		final int result = getConfig( getDeviceIndex(), configBlock );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		System.arraycopy( configBlock, 0, prevConfigBlock, 0, configBlock.length );
	}	// AnalogInputSubsystem()

	/*
	 * properties
	 */

	/**
	 * Prints the properties of this subsystem. Mainly useful for diagnostic purposes.
	 * @param stream the print stream where properties will be printed.
	 * @return The print stream.
	 */

	public PrintStream print( PrintStream stream ) {
		stream.println( 
			  "    Number of A/D channels: " + numChannels + "\n"
			+ "    Number of MUXed A/D channels: " + numMUXChannels
		);
		return stream;
	}	// print()

	/**
	 * Gets the number of primary analog input channels.
	 * @return Number of channels, numbered 0 to n-1.
	 */

	public int getNumChannels() {
		return numChannels;
	}	// getNumChannels()

	/**
	 * Gets the number of analog input channels available through an optional multiplexer.
	 * @return Number of channels, numbered 0 to n-1.
	 */

	public int getNumMUXChannels() {
		return numMUXChannels;
	}	// getNumMUXChannels()

	/**
	 * Tells if automatic calibration is possible with this device.
	 * @return <i>True</i> indicates that automatic calibration is available.
	 * @see #calibrate( boolean autoCal, boolean returnCalTable, String saveFileName ) calibrate()
	 */

	public boolean isAutoCalPresent( boolean force ) {
		if(
			force
			|| autoCalFeature == AUTO_CAL_UNKNOWN
		) {
			final int ERROR_NOT_SUPPORTED = 10;	// AIOUSB result code
			final int result = queryCal( getDeviceIndex() );
			if( result == USBDeviceManager.SUCCESS )
				autoCalFeature = AUTO_CAL_PRESENT;
			else if( result == ERROR_NOT_SUPPORTED )
				autoCalFeature = AUTO_CAL_NOT_PRESENT;
			else {
				/*
				 * an error occurred; leave auto. cal. feature unchanged
				 */
				throw new OperationFailedException( result );
			}	// else if( result ...
		}	// if( force ...
		return autoCalFeature == AUTO_CAL_PRESENT;
	}	// isAutoCalPresent()

	/*
	 * configuration
	 */

	/**
	 * Tells if the <i>read()</i> functions will discard the first A/D sample taken. Discarding the first sample may be useful in cases
	 * in which voltage "residue" from reading a different channel affects the channel currently being read.
	 * @return <i>False</i> indicates that no samples will be discarded; <i>true</i> indicates that the first sample will be discarded.
	 */

	public boolean isDiscardFirstSample() {
		return isDiscardFirstSample( getDeviceIndex() );
	}	// isDiscardFirstSample()

	/**
	 * Specifies whether the <i>read()</i> functions will discard the first A/D sample taken. Discarding the first sample may be useful
	 * in cases in which voltage "residue" from reading a different channel affects the channel currently being read.
	 * @param discard <i>false</i> indicates that no samples will be discarded; <i>true</i> indicates that the first sample will be discarded.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem setDiscardFirstSample( boolean discard ) {
		final int result = setDiscardFirstSample( getDeviceIndex(), discard );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// setDiscardFirstSample()

	/**
	 * Gets the current calibration mode.
	 * @return Current calibration mode (<i>CAL_MODE_NORMAL</i>, <i>CAL_MODE_GROUND</i> or <i>CAL_MODE_REFERENCE</i>).
	 * @see #setCalMode( int mode ) setCalMode()
	 */

	public int getCalMode() {
		validateConfigBlock();
		return configBlock[ REG_CAL_MODE ] & 0xff;
	}	// getCalMode()

	/**
	 * Sets the A/D calibration mode. If ground or reference mode is selected, only one A/D sample may be taken at a time.
	 * That means, one channel and no oversampling. Attempting to read more than one channel or use an oversample setting
	 * of more than zero will result in a timeout error because the device will not send more than one sample. In order to
	 * protect users from accidentally falling into this trap, the <i>read()</i> functions automatically and temporarily
	 * correct the scan parameters, restoring them when they complete.
	 * @param mode the calibration mode. May be one of:<br>
	 * <i>{@link #CAL_MODE_NORMAL}<br>
	 * {@link #CAL_MODE_GROUND}<br>
	 * {@link #CAL_MODE_REFERENCE}</i>
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setCalMode( int mode ) {
		validateConfigBlock();
		if(
			mode != CAL_MODE_NORMAL
			&& mode != CAL_MODE_GROUND
			&& mode != CAL_MODE_REFERENCE
		)
			throw new IllegalArgumentException( "Invalid cal. mode: " + mode );
		configBlock[ REG_CAL_MODE ] = ( byte ) mode;
		return writeConfig( false );
	}	// setCalMode()

	/**
	 * Gets the current trigger mode.
	 * @return Current trigger mode (bitwise OR of <i>TRIG_MODE_CTR0_EXT</i>, <i>TRIG_MODE_FALLING_EDGE</i>, <i>TRIG_MODE_SCAN</i>,
	 * <i>TRIG_MODE_EXTERNAL</i> or <i>TRIG_MODE_TIMER</i>).
	 * @see #setTriggerMode( int mode ) setTriggerMode()
	 */

	public int getTriggerMode() {
		validateConfigBlock();
		return configBlock[ REG_TRIG_MODE ] & TRIG_MODE_VALID_MASK;
	}	// getTriggerMode()

	/**
	 * Sets the trigger mode.
	 * @param mode a bitwise OR of these flags:<br>
	 * <i>{@link #TRIG_MODE_CTR0_EXT}<br>
	 * {@link #TRIG_MODE_FALLING_EDGE}<br>
	 * {@link #TRIG_MODE_SCAN}<br>
	 * {@link #TRIG_MODE_EXTERNAL}<br>
	 * {@link #TRIG_MODE_TIMER}</i>
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setTriggerMode( int mode ) {
		validateConfigBlock();
		if( ( mode & ~TRIG_MODE_VALID_MASK ) != 0 )
			throw new IllegalArgumentException( "Invalid trigger mode: " + mode );
		configBlock[ REG_TRIG_MODE ] = ( byte ) mode;
		return writeConfig( false );
	}	// setTriggerMode()

	/**
	 * Gets the current gain code for <i>channel</i>.
	 * @param channel the channel for which to obtain the current gain code.
	 * @return Current gain code for <i>channel</i>.
	 * @see #setGainCode( int channel, int gainCode ) setGainCode()
	 * @throws IllegalArgumentException
	 */

	public int getGainCode( int channel ) {
		validateConfigBlock();
		if(
			channel < 0
			|| channel >= numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid channel: " + channel );
		return configBlock[ REG_GAIN_CODE + channel % NUM_GAIN_CODE_REGISTERS ] & GAIN_CODE_VALID_MASK;
	}	// getGainCode()

	/**
	 * Gets the current gain code for multiple A/D channels.
	 * @param startChannel the first channel for which to obtain the current gain code.
	 * @param numChannels the number of channels for which to obtain the current gain code.
	 * @return Array containing the current gain codes for the specified channels.
	 * @see #setGainCode( int startChannel, int[] gainCode ) setGainCode()
	 * @throws IllegalArgumentException
	 */

	public int[] getGainCode( int startChannel, int numChannels ) {
		validateConfigBlock();
		if(
			numChannels < 1
			|| startChannel < 0
			|| startChannel + numChannels > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid start channel: " + startChannel
				+ ", or number of channels: " + numChannels );
		int[] gainCode = new int[ numChannels ];
		for( int index = 0; index < gainCode.length; index++ ) {
			gainCode[ index ]
				= configBlock[ REG_GAIN_CODE + ( startChannel + index ) % NUM_GAIN_CODE_REGISTERS ]
				& GAIN_CODE_VALID_MASK;
		}	// for( int index ...
		return gainCode;
	}	// getGainCode()

	/**
	 * Sets the gain code for a single A/D channel.
	 * @param channel the channel for which to set the gain code.
	 * @param gainCode the gain code (voltage range) for the channel. May be one of:<br>
	 * <i>{@link #GAIN_CODE_0_10V}<br>
	 * {@link #GAIN_CODE_10V}<br>
	 * {@link #GAIN_CODE_0_5V}<br>
	 * {@link #GAIN_CODE_5V}<br>
	 * {@link #GAIN_CODE_0_2V}<br>
	 * {@link #GAIN_CODE_2V}<br>
	 * {@link #GAIN_CODE_0_1V}<br>
	 * {@link #GAIN_CODE_1V}</i>
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogInputSubsystem setGainCode( int channel, int gainCode ) {
		int[] gainCodes = { gainCode };
		return setGainCode( channel, gainCodes );
	}	// setGainCode()

	/**
	 * Sets the gain code for multiple A/D channels.
	 * @param startChannel the first channel for which to set the gain code.
	 * @param gainCode an array of gain codes, one per channel <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setGainCode( int startChannel, int[] gainCode ) {
		validateConfigBlock();
		if(
			gainCode == null
			|| gainCode.length < 1
			|| startChannel < 0
			|| startChannel + gainCode.length > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid gain code array or start channel: " + startChannel );
		for( int index = 0; index < gainCode.length; index++ ) {
			if( ( gainCode[ index ] & ~GAIN_CODE_VALID_MASK ) != 0 )
				throw new IllegalArgumentException( "Invalid gain code: " + gainCode[ index ] );
			final int reg = REG_GAIN_CODE + ( startChannel + index ) % NUM_GAIN_CODE_REGISTERS;
			configBlock[ reg ] = ( byte ) ( ( configBlock[ reg ] & ~GAIN_CODE_VALID_MASK ) | gainCode[ index ] );
		}	// for( int index ...
		return writeConfig( false );
	}	// setGainCode()

	/**
	 * Tells if <i>channel</i> is configured for single-ended or differential mode.
	 * @param channel the channel for which to obtain the current differential mode.
	 * @return Current gain code for <i>channel</i>.
	 * @return <i>False</i> indicates single-ended mode; <i>true</i> indicates differential mode.
	 * @see #setDifferentialMode( int channel, boolean differentialMode ) setDifferentialMode()
	 * @throws IllegalArgumentException
	 */

	public boolean isDifferentialMode( int channel ) {
		validateConfigBlock();
		if(
			channel < 0
			|| channel >= numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid channel: " + channel );
		return ( configBlock[ REG_GAIN_CODE + channel % NUM_GAIN_CODE_REGISTERS ] & DIFFERENTIAL_MODE ) != 0;
	}	// isDifferentialMode()

	/**
	 * Tells if multiple A/D channels are configured for single-ended or differential mode.
	 * @param startChannel the first channel for which to obtain the current differential mode.
	 * @param numChannels the number of channels for which to obtain the current differential mode.
	 * @return Array containing the current differential modes for the specified channels. <i>False</i> indicates
	 * channel is configured for single-ended mode and <i>true</i> indicates channel is configured for differential mode.
	 * @see #setDifferentialMode( int startChannel, boolean[] differentialMode ) setDifferentialMode()
	 * @throws IllegalArgumentException
	 */

	public boolean[] isDifferentialMode( int startChannel, int numChannels ) {
		validateConfigBlock();
		if(
			numChannels < 1
			|| startChannel < 0
			|| startChannel + numChannels > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid start channel: " + startChannel
				+ ", or number of channels: " + numChannels );
		boolean[] differentialMode = new boolean[ numChannels ];
		for( int index = 0; index < differentialMode.length; index++ ) {
			differentialMode[ index ]
				= ( configBlock[ REG_GAIN_CODE + ( startChannel + index ) % NUM_GAIN_CODE_REGISTERS ] & DIFFERENTIAL_MODE ) != 0;
		}	// for( int index ...
		return differentialMode;
	}	// isDifferentialMode()

	/**
	 * Sets a single A/D channel to differential or single-ended mode.
	 * @param channel the channel for which to set differential or single-ended mode.
	 * @param differentialMode <i>false</i> selects single-ended mode; <i>true</i> selects differential mode.
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogInputSubsystem setDifferentialMode( int channel, boolean differentialMode ) {
		boolean[] differentialModes = { differentialMode };
		return setDifferentialMode( channel, differentialModes );
	}	// setDifferentialMode()

	/**
	 * Sets multiple A/D channels to differential or single-ended mode.
	 * @param startChannel the first channel for which to set differential or single-ended mode.
	 * @param differentialMode an array of mode selectors, one per channel. For each element in the array,
	 * <i>false</i> selects single-ended mode for that channel and <i>true</i> selects differential mode.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setDifferentialMode( int startChannel, boolean[] differentialMode ) {
		validateConfigBlock();
		if(
			differentialMode == null
			|| differentialMode.length < 1
			|| startChannel < 0
			|| startChannel + differentialMode.length > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid differential mode array or start channel: " + startChannel );
		for( int index = 0; index < differentialMode.length; index++ ) {
			final int reg = REG_GAIN_CODE + ( startChannel + index ) % NUM_GAIN_CODE_REGISTERS;
			if( differentialMode[ index ] )
				configBlock[ reg ] |= DIFFERENTIAL_MODE;
			else
				configBlock[ reg ] &= ~DIFFERENTIAL_MODE;
		}	// for( int index ...
		return writeConfig( false );
	}	// setDifferentialMode()

	/**
	 * Sets the gain code and differential mode for a single A/D channel.
	 * @param channel the channel for which to set the gain code.
	 * @param gainCode the gain code (voltage range) for the channel <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @param differentialMode <i>false</i> selects single-ended mode; <i>true</i> selects differential mode.
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogInputSubsystem setGainCodeAndDiffMode( int channel, int gainCode, boolean differentialMode ) {
		int[] gainCodes = { gainCode };
		boolean[] differentialModes = { differentialMode };
		return setGainCodeAndDiffMode( channel, gainCodes, differentialModes );
	}	// setGainCodeAndDiffMode()

	/**
	 * Sets the gain code and differential mode for multiple A/D channels.
	 * @param startChannel the first channel for which to set the gain code and differential mode.
	 * @param gainCode an array of gain codes, one per channel <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @param differentialMode an array of mode selectors, one per channel. For each element in the array,
	 * <i>false</i> selects single-ended mode for that channel and <i>true</i> selects differential mode.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setGainCodeAndDiffMode( int startChannel, int[] gainCode, boolean[] differentialMode ) {
		validateConfigBlock();
		if(
			gainCode == null
			|| gainCode.length < 1
			|| differentialMode == null
			|| differentialMode.length != gainCode.length
			|| startChannel < 0
			|| startChannel + gainCode.length > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid gain code array, differential mode array or start channel: " + startChannel );
		for( int index = 0; index < gainCode.length; index++ ) {
			if( ( gainCode[ index ] & ~GAIN_CODE_VALID_MASK ) != 0 )
				throw new IllegalArgumentException( "Invalid gain code: " + gainCode[ index ] );
			final int reg = REG_GAIN_CODE + ( startChannel + index ) % NUM_GAIN_CODE_REGISTERS;
			configBlock[ reg ] = ( byte ) ( ( configBlock[ reg ] & ~GAIN_CODE_VALID_MASK ) | gainCode[ index ] );
			if( differentialMode[ index ] )
				configBlock[ reg ] |= DIFFERENTIAL_MODE;
			else
				configBlock[ reg ] &= ~DIFFERENTIAL_MODE;
		}	// for( int index ...
		return writeConfig( false );
	}	// setGainCodeAndDiffMode()

	/**
	 * Sets all the A/D channels to the same gain code and differential mode.
	 * @param gainCode the gain code (voltage range) for the channels <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @param differentialMode <i>false</i> selects single-ended mode; <i>true</i> selects differential mode.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setGainCodeAndDiffMode( int gainCode, boolean differentialMode ) {
		validateConfigBlock();
		if( ( gainCode & ~GAIN_CODE_VALID_MASK ) != 0 )
			throw new IllegalArgumentException( "Invalid gain code: " + gainCode );
		for( int index = 0; index < NUM_GAIN_CODE_REGISTERS; index++ ) {
			final int reg = REG_GAIN_CODE + index;
			configBlock[ reg ] = ( byte ) ( ( configBlock[ reg ] & ~GAIN_CODE_VALID_MASK ) | gainCode );
			if( differentialMode )
				configBlock[ reg ] |= DIFFERENTIAL_MODE;
			else
				configBlock[ reg ] &= ~DIFFERENTIAL_MODE;
		}	// for( int index ...
		return writeConfig( false );
	}	// setGainCodeAndDiffMode()

	/**
	 * Gets the current number of over-samples.
	 * @return Current number of over-samples (0-255).
	 * @see #setOversample( int oversample ) setOversample()
	 */

	public int getOversample() {
		validateConfigBlock();
		return configBlock[ REG_OVERSAMPLE ] & 0xff;
	}	// getOversample()

	/**
	 * Sets the number of over-samples for all A/D channels.
	 * @param oversample number of over-samples (0-255).
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogInputSubsystem setOversample( int oversample ) {
		validateConfigBlock();
		if(
			oversample < 0
			|| oversample > MAX_OVERSAMPLE
		)
			throw new IllegalArgumentException( "Invalid oversample: " + oversample );
		configBlock[ REG_OVERSAMPLE ] = ( byte ) oversample;
		return writeConfig( false );
	}	// setOversample()

	/**
	 * Loads a calibration table from a file into the A/D.
	 * @param fileName the name of a file containing the calibration table. A calibration table must consist
	 * of exactly 65,536 16-bit unsigned integers
	 * <i>(see {@link #calibrate( boolean autoCal, boolean returnCalTable, String saveFileName ) calibrate()})</i>.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem setCalibrationTable( String fileName ) {
		if(
			fileName == null
			|| fileName.isEmpty()
		)
			throw new IllegalArgumentException( "Invalid file name" );
		final int result = loadCalTable( getDeviceIndex(), fileName );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// setCalibrationTable()

	/**
	 * Sets the calibration table in the A/D to the contents of <i>calTable</i>.
	 * @param calTable the calibration table to load. A calibration table must consist
	 * of exactly 65,536 16-bit unsigned integers
	 * <i>(see {@link #calibrate( boolean autoCal, boolean returnCalTable, String saveFileName ) calibrate()})</i>.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem setCalibrationTable( char[] calTable ) {
		if(
			calTable == null
			|| calTable.length != CAL_TABLE_WORDS
		)
			throw new IllegalArgumentException( "Invalid cal. table" );
		final int result = setCalTable( getDeviceIndex(), calTable );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// setCalibrationTable()

	/**
	 * Sets the streaming block size.
	 * @param blockSize the streaming block size you wish to set. This will be rounded up to the
	 * next multiple of 512.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem setStreamingBlockSize( int blockSize ) {
		if( blockSize < 0 )
			throw new IllegalArgumentException( "Invalid block size: " + blockSize );
		final int result = setStreamingBlockSize( getDeviceIndex(), blockSize );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// setStreamingBlockSize()

	/**
	 * Sets the clock frequency for timer-driven bulk reads <i>(see
	 * {@link #readBulkStart( int startChannel, int numChannels, int numSamples ) readBulkStart()})</i>.
	 * @param clockHz the frequency at which to take the samples (in Hertz).
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogInputSubsystem setClock( double clockHz ) {
		parent.setMiscClock( clockHz );
		return this;
	}	// setClock()

	/**
	 * Calibrates the A/D, generating either a default table or using the internal voltage references
	 * to generate a calibration table.
	 * @param autoCal <i>false</i> generates a default (uncalibrated) table; <i>true</i> uses the internal voltage
	 * references to automatically calibrate the A/D.
	 * @param returnCalTable <i>false</i> causes <i>calibrate()</i> to return <i>null</i>; <i>true</i> causes it to
	 * return the generated calibration table.
	 * @param saveFileName the name of the file in which to save the generated calibration table. If <i>null</i> or
	 * empty, the generated calibration table is not saved to a file.
	 * @return If <i>returnCalTable</i> is <i>false</i>, <i>null</i> is returned. Otherwise an array of 65,536 16-bit
	 * unsigned integers representing the generated calibration table is returned.
	 * @throws OperationFailedException
	 */

	public char[] calibrate( boolean autoCal, boolean returnCalTable, String saveFileName ) {
		char[] calTable = null;
		if( returnCalTable )
			calTable = new char[ CAL_TABLE_WORDS ];
		final int result = internalCal( getDeviceIndex(), autoCal, calTable, saveFileName );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return calTable;
	}	// calibrate()

	/**
	 * Permits the A/D to be calibrated using an external voltage source. The proper way to use this function is to configure
	 * the A/D with a default calibration table (such as by calling
	 * <i>{@link #calibrate( boolean autoCal, boolean returnCalTable, String saveFileName ) calibrate( false ... )}</i>). Then
	 * inject a series of voltages into one of the A/D input channels, recording the count values reported by the A/D (by calling
	 * <i>{@link #read( int channel ) read()}</i>). It's also a good idea to enable oversampling while recording these values in order
	 * to obtain the most stable readings. Alternatively, since <i>points</i> is an array of <i>double</i> values, you can obtain
	 * individual A/D count measurements and average them yourself, producing a <i>double</i> average, and put that value into
	 * the <i>points</i> array.<br><br>
	 * The <i>points</i> array consists of voltage-count pairs; <i>points[0]</i> is the first input voltage; <i>points[1]</i> is
	 * the corresponding count value measured by the A/D; <i>points[2]</i> and <i>points[3]</i> contain the second pair of
	 * voltage-count values; and so on. You can provide any number of pairs, although more than a few dozen is probably overkill,
	 * not to mention would take a lot of effort to acquire.<br><br>
	 * This calibration procedure uses the current gain A/D setting for channel 0, so it must be the same as that used to collect
	 * the measured A/D counts. It's recommended that all the channels be set to the same gain, the one that will be used during
	 * normal operation. The calibration is gain dependent, so switching the gain after calibrating may introduce slight offset or
	 * gain changes. So for best results, the A/D should be calibrated on the same gain setting that will be used during normal
	 * operation. You can create any number of calibration tables. If your application needs to switch between ranges, you may wish
	 * to create a separate calibration table for each range your application will use. Then when switching to a different range,
	 * the application can load the appropriate calibration table.<br><br>
	 * Although calibrating in this manner does take some effort, it produces the best results, eliminating all sources of error
	 * from the input pins onward. Furthermore, the calibration table can be saved to a file and reloaded into the A/D, ensuring consistency.
	 * @param points array of voltage-count pairs to calibrate the A/D with.
	 * @param returnCalTable <i>false</i> causes <i>calibrate()</i> to return <i>null</i>; <i>true</i> causes it to
	 * return the generated calibration table.
	 * @param saveFileName the name of the file in which to save the generated calibration table. If <i>null</i> or
	 * empty, the generated calibration table is not saved to a file.
	 * @return If <i>returnCalTable</i> is <i>false</i>, <i>null</i> is returned. Otherwise an array of 65,536 16-bit
	 * unsigned integers representing the generated calibration table is returned.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public char[] calibrate( double[] points, boolean returnCalTable, String saveFileName ) {
		if(
			points == null
			|| points.length < 2
			|| ( points.length & 1 ) != 0
		)
			throw new IllegalArgumentException( "Invalid points array" );
		char[] calTable = null;
		if( returnCalTable )
			calTable = new char[ CAL_TABLE_WORDS ];
		final int result = externalCal( getDeviceIndex(), points, calTable, saveFileName );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return calTable;
	}	// calibrate()

	/**
	 * Reads the A/D count value from a single channel.
	 * @param channel the channel to read.
	 * @return A/D counts (0-65,535). The meaning of these counts depends on the current A/D range of the channel
	 * <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. The count value may be converted to
	 * a voltage value using <i>{@link #countsToVolts( int channel, char counts ) countsToVolts()}</i>.
	 */

	public char read( int channel ) {
		return read( channel, 1 )[ 0 ];
	}	// read()

	/**
	 * Reads the A/D count values from multiple channels.
	 * @param startChannel the first channel to read.
	 * @param numChannels the number of channels to read.
	 * @return An array of A/D counts (0-65,535), one per channel read. The meaning of these counts depend on the
	 * current A/D range of each channel <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * The array of count values may be converted to an array of voltage values using
	 * <i>{@link #countsToVolts( int startChannel, char[] counts ) countsToVolts()}</i>.
	 * @throws OperationFailedException
	 */

	public char[] read( int startChannel, int numChannels ) {
		/*
		 * make copy of configuration block and restore it when done
		 */
		byte[] origConfigBlock = Arrays.copyOf( configBlock, configBlock.length );

		/*
		 * if calibration mode is enabled, then don't permit more than one
		 * sample to be taken; otherwise "bad" things will happen
		 */
		int calMode = getCalMode();
		if(
			calMode == CAL_MODE_GROUND
			|| calMode == CAL_MODE_REFERENCE
		) {
			setDiscardFirstSample( false );
			setOversample( 0 );
			numChannels = 1;
		}	// if( calMode ...
		setScanRange( startChannel, numChannels );
		char[] counts = new char[ numChannels ];
		final int result = getScan( getDeviceIndex(), counts );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );

		/*
		 * restore original configuration block
		 */
		configBlock = origConfigBlock;
		writeConfig( false );

		return counts;
	}	// read()

	/**
	 * Reads the voltage from a single channel.
	 * @param channel the channel to read.
	 * @return A voltage value, limited to the current A/D range of the channel <i>(see
	 * {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. The voltage value may be converted to a
	 * count value using <i>{@link #voltsToCounts( int channel, double volts ) voltsToCounts()}</i>.
	 */

	public double readVolts( int channel ) {
		return readVolts( channel, 1 )[ 0 ];
	}	// readVolts()

	/**
	 * Reads the voltage from multiple channels.
	 * @param startChannel the first channel to read.
	 * @param numChannels the number of channels to read.
	 * @return An array of voltage values, one per channel read, each limited to the current A/D range of each channel
	 * <i>(see {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. The array of voltage values may be
	 * converted to an array of count values using <i>{@link #voltsToCounts( int startChannel, double[] volts ) voltsToCounts()}</i>.
	 */

	public double[] readVolts( int startChannel, int numChannels ) {
		return countsToVolts( startChannel, read( startChannel, numChannels ) );
	}	// readVolts()

	/**
	 * Starts a large A/D acquisition process in a background thread and returns immediately. The status of the
	 * acquisition process can be monitored using <i>{@link #readBulkSamplesAvailable readBulkSamplesAvailable()}</i>, which
	 * returns the number of samples available to be retrieved by <i>{@link #readBulkNext( int numSamples ) readBulkNext()}</i>.
	 * When the last of the data has been retrieved using <i>readBulkNext()</i>, the bulk acquisition process is
	 * automatically terminated and becomes ready to be used again.<br><br>
	 * <i>While a bulk acquisition process is in progress, no functions of the device other than readBulkSamplesAvailable()
	 * or readBulkNext() should be used.</i><br><br>
	 * This example shows the proper way to configure the device for a large A/D acquisition process using the internal timer.
	 * <pre>device.adc()
	 *   .setStreamingBlockSize( 100000 )
	 *   .setCalMode( AnalogInputSubsystem.CAL_MODE_NORMAL )
	 *   .setTriggerMode( AnalogInputSubsystem.TRIG_MODE_SCAN | AnalogInputSubsystem.TRIG_MODE_TIMER )
	 *   .setClock( 100000 )
	 *   .readBulkStart( 1, 1, numSamples );
	 * do {
	 *   char[] data = device.adc().readBulkNext( 20000 );
	 *   ... do something with data ...
	 * } while( ...more data is available... );</pre>
	 * @param startChannel the first channel to read.
	 * @param numChannels the number of channels to read.
	 * @param numSamples the total number of <i>samples</i> to read.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogInputSubsystem readBulkStart( int startChannel, int numChannels, int numSamples ) {
		if( numSamples < 1 )
			throw new IllegalArgumentException( "Invalid number of samples: " + numSamples );
		setScanRange( startChannel, numChannels );
		final int result = readBulkStart( getDeviceIndex(), numSamples );
		if( result == USBDeviceManager.SUCCESS ) {
			readBulkSamplesRequested = numSamples;
			readBulkSamplesRetrieved = 0;
		} else {
			readBulkSamplesRequested = readBulkSamplesRetrieved = 0;
			throw new OperationFailedException( result );
		}	// if( result ...
		return this;
	}	// readBulkStart()

	/**
	 * Gets the number of samples available to be retrieved during a bulk acquisition process initiated by
	 * <i>{@link #readBulkStart( int startChannel, int numChannels, int numSamples ) readBulkStart()}</i>.
	 * @return The number of samples available.
	 * @throws OperationFailedException
	 */

	public int readBulkSamplesAvailable() {
		int samplesAvailable = 0;
		if( readBulkSamplesRequested >= 0 ) {
			int[] samplesRemaining = new int[ 1 ];
			final int result = readBulkSamplesRemaining( getDeviceIndex(), samplesRemaining );
			if( result != USBDeviceManager.SUCCESS )
				throw new OperationFailedException( result );
			samplesAvailable = readBulkSamplesRequested - readBulkSamplesRetrieved - samplesRemaining[ 0 ];
			if( samplesAvailable < 0 )
				samplesAvailable = 0;
		}	// if( readBulkSamplesRequested ...
		return samplesAvailable;
	}	// readBulkSamplesAvailable()

	/**
	 * Retrieves the next set of samples acquired during a bulk acquisition process initiated by
	 * <i>{@link #readBulkStart( int startChannel, int numChannels, int numSamples ) readBulkStart()}</i>.
	 * @param numSamples the number of samples to retrieve.
	 * @return An array containing the number of samples requested, or all that are available. If fewer samples
	 * are available than are requested, only the samples available are returned. If zero samples are available,
	 * a zero-length array is returned.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public char[] readBulkNext( int numSamples ) {
		if( numSamples < 1 )
			throw new IllegalArgumentException( "Invalid number of samples: " + numSamples );
		char[] counts = new char[ 0 ];
		if( readBulkSamplesRequested >= 0 ) {
			int samplesAvailable = readBulkSamplesAvailable();
			if( samplesAvailable > 0 ) {
				if( numSamples > samplesAvailable )
					numSamples = samplesAvailable;
				counts = new char[ numSamples ];
				int result = readBulkNext( getDeviceIndex(), readBulkSamplesRetrieved, counts );
				if( result == USBDeviceManager.SUCCESS ) {
					if( ( samplesAvailable -= numSamples ) <= 0 ) {
						/*
						 * no more data remains to be retrieved; finish bulk read procedure
						 */
						result = readBulkFinish( getDeviceIndex() );
						readBulkSamplesRequested = readBulkSamplesRetrieved = 0;
						if( result != USBDeviceManager.SUCCESS )
							throw new OperationFailedException( result );
					} else
						readBulkSamplesRetrieved += numSamples;
				} else
					throw new OperationFailedException( result );
			}	// if( samplesAvailable ...
		}	// if( readBulkSamplesRequested ...
		return counts;
	}	// readBulkNext()

	/**
	 * Clears the streaming FIFO, using one of several different methods.
	 * @param method the method to use when clearing the FIFO. May be one of:<br>
	 * <i>{@link USBDevice#CLEAR_FIFO_METHOD_IMMEDIATE}<br>
	 * {@link USBDevice#CLEAR_FIFO_METHOD_AUTO}<br>
	 * {@link USBDevice#CLEAR_FIFO_METHOD_IMMEDIATE_AND_ABORT}<br>
	 * {@link USBDevice#CLEAR_FIFO_METHOD_WAIT}</i>
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogInputSubsystem clearFIFO( int method ) {
		parent.clearFIFO( method );
		return this;
	}	// clearFIFO()

	/*
	 * utilities
	 */

	/**
	 * Converts a single A/D count value to volts, based on the current gain setting for the specified channel.
	 * Be careful to ensure that the count value was actually obtained from the specified channel and that the gain hasn't
	 * changed since the count value was obtained.
	 * @param channel the channel number to use for converting counts to volts.
	 * @param counts the count value to convert to volts.
	 * @return A voltage value calculated using the <i>current</i> A/D range of the channel <i>(see
	 * {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @throws IllegalArgumentException
	 */

	public double countsToVolts( int channel, char counts ) {
		if(
			channel < 0
			|| channel >= numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid channel: " + channel );
		return countsToVolts( getDeviceIndex(), channel, counts );
	}	// countsToVolts()

	/**
	 * Converts an array of A/D count values to an array of voltage values, based on the current gain setting for each of
	 * the specified channels. Be careful to ensure that the count values were actually obtained from the specified channels
	 * and that the gains havn't changed since the count values were obtained.
	 * @param startChannel the first channel number to use for converting counts to volts.
	 * @param counts the count values to convert to volts.
	 * @return An array of voltage values calculated using the <i>current</i> A/D range of each of the channels <i>(see
	 * {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. The array returned has the same number of
	 * elements as <i>counts</i>.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public double[] countsToVolts( int startChannel, char[] counts ) {
		if(
			counts == null
			|| counts.length < 1
			|| startChannel < 0
			|| startChannel + counts.length > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid counts array or start channel: " + startChannel );
		double[] volts = new double[ counts.length ];
		final int result = countsToVolts( getDeviceIndex(), startChannel, counts, volts );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return volts;
	}	// countsToVolts()

	/**
	 * Converts a single voltage value to A/D counts, based on the current gain setting for the specified channel.
	 * Be careful to ensure that the voltage value was actually obtained from the specified channel and that the gain hasn't
	 * changed since the voltage value was obtained.
	 * @param channel the channel number to use for converting volts to counts.
	 * @param volts the voltage value to convert to counts.
	 * @return A count value calculated using the <i>current</i> A/D range of the channel <i>(see
	 * {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>.
	 * @throws IllegalArgumentException
	 */

	public char voltsToCounts( int channel, double volts ) {
		if(
			channel < 0
			|| channel >= numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid channel: " + channel );
		return voltsToCounts( getDeviceIndex(), channel, volts );
	}	// voltsToCounts()

	/**
	 * Converts an array of voltage values to an array of A/D count values, based on the current gain setting for each of
	 * the specified channels. Be careful to ensure that the voltage values were actually obtained from the specified channels
	 * and that the gains havn't changed since the voltage values were obtained.
	 * @param startChannel the first channel number to use for converting volts to counts.
	 * @param volts the voltage values to convert to counts.
	 * @return An array of count values calculated using the <i>current</i> A/D range of each of the channels <i>(see
	 * {@link #setGainCode( int channel, int gainCode ) setGainCode()})</i>. The array returned has the same number of
	 * elements as <i>volts</i>.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public char[] voltsToCounts( int startChannel, double[] volts ) {
		if(
			volts == null
			|| volts.length < 1
			|| startChannel < 0
			|| startChannel + volts.length > numMUXChannels
		)
			throw new IllegalArgumentException( "Invalid volts array or start channel: " + startChannel );
		char[] counts = new char[ volts.length ];
		final int result = voltsToCounts( getDeviceIndex(), startChannel, volts, counts );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return counts;
	}	// voltsToCounts()

	// }}}

}	// class AnalogInputSubsystem

/* end of file */
