/*
 * $RCSfile: AnalogOutputSubsystem.java,v $
 * $Date: 2009/11/29 17:11:48 $
 * $Revision: 1.9 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 */

package com.acces.aiousb;

// {{{ imports
import java.io.*;
import java.util.*;
import java.lang.Math;
// }}}

/**
 * Class AnalogOutputSubsystem represents the analog output subsystem of a device. One accesses
 * this analog output subsystem through its parent object, typically through a method such as
 * <i>dac() (see {@link USB_AO16_Family#dac()})</i>.
 */

public class AnalogOutputSubsystem extends DeviceSubsystem {

	// {{{ public constants

	/*
	 * range codes passed to setRange()
	 */
	/** Unipolar, 0-5 volt range <i>(see {@link #setRange( int range ) setRange()})</i>. */
	public static final int DAC_RANGE_0_5V		= 0;	// 0-5V
	/** Bipolar, -5 to +5 volt range. */
	public static final int DAC_RANGE_5V		= 1;	// +/-5V
	/** Unipolar, 0-10 volt range. */
	public static final int DAC_RANGE_0_10V		= 2;	// 0-10V
	/** Bipolar, -10 to +10 volt range. */
	public static final int DAC_RANGE_10V		= 3;	// +/-10V

	/** Maximum number of counts D/A can output. */
	public static final int MAX_COUNTS = 65535;

	// }}}

	// {{{ protected members

	protected static final int NUM_RANGES = 4;
	protected int range;						// current D/A range
	protected int numChannels;					// number of D/A channels
	protected static final double daRange[][] = {
		//  min, range
		  {   0,  5 }							// DAC_RANGE_0_5V
		, {  -5, 10 }							// DAC_RANGE_5V
		, {   0, 10 }							// DAC_RANGE_0_10V
		, { -10, 20 }							// DAC_RANGE_10V
	};	// daRange[]

	// }}}

	// {{{ protected methods

	protected native int getNumChannels( int deviceIndex );
	protected native int setBoardRange( int deviceIndex, int range );
		// => unsigned long DACSetBoardRange( unsigned long DeviceIndex, unsigned long RangeCode )
	protected native int write( int deviceIndex, int channel, char counts );
		// => unsigned long DACDirect( unsigned long DeviceIndex, unsigned short Channel, unsigned short Value )
	protected native int write( int deviceIndex, char[] points );
		// => unsigned long DACMultiDirect( unsigned long DeviceIndex, unsigned short *pDACData, unsigned long DACDataCount )
	// }}}

	// {{{ public methods

	/**
	 * Constructor for analog output subsystem.
	 * @param parent the parent USB device object that this subsystem is a part of.
	 */

	public AnalogOutputSubsystem( USBDevice parent ) {
		super( parent );
		range = DAC_RANGE_0_5V;
		numChannels = getNumChannels( getDeviceIndex() );
		assert numChannels > 0;
		setRange( DAC_RANGE_0_5V );
	}	// AnalogOutputSubsystem()

	/*
	 * properties
	 */

	/**
	 * Prints the properties of this subsystem. Mainly useful for diagnostic purposes.
	 * @param stream the print stream where properties will be printed.
	 * @return The print stream.
	 */

	public PrintStream print( PrintStream stream ) {
		stream.println( "    Number of D/A channels: " + numChannels );
		return stream;
	}	// print()

	/**
	 * Gets the number of analog output channels.
	 * @return Number of channels, numbered 0 to n-1.
	 */

	public int getNumChannels() {
		return numChannels;
	}	// getNumChannels()

	/*
	 * configuration
	 */

	/**
	 * Gets the current voltage range of the D/A outputs.
	 * @return Current voltage range.
	 * @see #setRange( int range ) setRange()
	 */

	public int getRange() {
		return range;
	}	// getRange()

	/**
	 * Sets the voltage range of the D/A outputs.
	 * @param range the voltage range to select. May be one of:<br>
	 * <i>{@link #DAC_RANGE_0_5V}<br>
	 * {@link #DAC_RANGE_5V}<br>
	 * {@link #DAC_RANGE_0_10V}<br>
	 * {@link #DAC_RANGE_10V}</i>
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogOutputSubsystem setRange( int range ) {
		if(
			range < DAC_RANGE_0_5V
			|| range > DAC_RANGE_10V
		)
			throw new IllegalArgumentException( "Invalid range: " + range );
		final int result = setBoardRange( getDeviceIndex(), range );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		this.range = range;
		return this;
	}	// setRange()

	/*
	 * operations
	 */

	/**
	 * Writes a count value to a D/A channel.
	 * @param channel the channel to write to.
	 * @param counts the D/A count value to output. The number of bits of resolution for the D/A outputs varies from
	 * model to model, however it's usually 12 or 16 bits. Moreover, some of the 12-bit models actually accept a
	 * 16-bit value and simply truncate the least significant 4 bits. Consult the manual for the specific device to
	 * determine the range of D/A values the device will accept. In general, 12-bit devices accept a count range of
	 * 0-0xfff, and 16-bit devices accept a count range of 0-0xffff.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogOutputSubsystem write( int channel, char counts ) {
		if(
			channel < 0
			|| channel >= numChannels
			|| counts < 0
			|| counts > MAX_COUNTS
		)
			throw new IllegalArgumentException( "Invalid channel: " + channel + ", or counts: " + counts );
		final int result = write( getDeviceIndex(), channel, counts );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// write()

	/**
	 * Writes a block of count values to one or more D/A channels.
	 * @param points an array of 16-bit integers representing channel-count pairs. The first integer of each pair
	 * is the D/A channel number and the second integer is the D/A count value to output to the specified channel.
	 * Refer to <i>{@link #write( int channel, char counts ) write()}</i> for an explanation of the channel addressing
	 * and count values.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 * @throws OperationFailedException
	 */

	public AnalogOutputSubsystem write( char[] points ) {
		if(
			points == null
			|| points.length < 2
			|| ( points.length & 1 ) != 0
		)
			throw new IllegalArgumentException( "Invalid points array" );

		/*
		 * check all channel/count pairs
		 */
		for( int index = 0; index < points.length; index += 2 ) {
			if(
				points[ index ] < 0
				|| points[ index ] >= numChannels
				|| points[ index + 1 ] < 0
				|| points[ index + 1 ] > MAX_COUNTS
			)
				throw new IllegalArgumentException( "Invalid channel: " + points[ index ] + ", or counts: " + points[ index + 1 ] );
		}	// for( int index ...

		final int result = write( getDeviceIndex(), points );
		if( result != USBDeviceManager.SUCCESS )
			throw new OperationFailedException( result );
		return this;
	}	// write()

	/**
	 * Writes a voltage value to a D/A channel.
	 * @param channel the channel to write to.
	 * @param volts the voltage value to output.
	 * @return This subsystem, useful for chaining together multiple operations.
	 */

	public AnalogOutputSubsystem writeVolts( int channel, double volts ) {
		return write( channel, voltsToCounts( volts ) );
	}	// writeVolts()

	/**
	 * Writes a block of voltage values to one or more D/A channels.
	 * @param points an array of {@link OutputVoltagePoint} points representing channel-voltage pairs.
	 * @return This subsystem, useful for chaining together multiple operations.
	 * @throws IllegalArgumentException
	 */

	public AnalogOutputSubsystem writeVolts( OutputVoltagePoint[] points ) {
		if(
			points == null
			|| points.length < 1
		)
			throw new IllegalArgumentException( "Invalid points array" );

		/*
		 * convert channel/voltage pairs to channel/count pairs
		 */
		char[] countPoints = new char[ points.length * 2 ];
		for( int index = 0; index < points.length; index++ ) {
			if(
				points[ index ].channel >= 0
				&& points[ index ].channel < numChannels
			) {
				countPoints[ index * 2 ] = ( char ) points[ index ].channel;
				countPoints[ index * 2 + 1 ] = voltsToCounts( points[ index ].volts );
			} else
				throw new IllegalArgumentException( "Invalid channel: " + points[ index ].channel );
		}	// for( int index ...

		return write( countPoints );
	}	// writeVolts()

	/**
	 * Converts a single D/A count value to volts, based on the current range setting.
	 * @param counts the count value to convert to volts.
	 * @return A voltage value calculated using the <i>current</i> D/A range.
	 * <i>(see {@link #setRange( int range ) setRange()})</i>.
	 */

	public double countsToVolts( char counts ) {
		assert range >= DAC_RANGE_0_5V
			&& range <= DAC_RANGE_10V;
		return ( ( double ) counts / MAX_COUNTS ) * daRange[ range ][ 1 ] + daRange[ range ][ 0 ];
	}	// countsToVolts()

	/**
	 * Converts an array of D/A count values to an array of voltage values, based on the current range setting.
	 * @param counts the array of count values to convert to volts.
	 * @return An array of voltage values calculated using the <i>current</i> D/A range.
	 * <i>(see {@link #setRange( int range ) setRange()})</i>.
	 * @throws IllegalArgumentException
	 */

	public double[] countsToVolts( char[] counts ) {
		if(
			counts == null
			|| counts.length < 1
		)
			throw new IllegalArgumentException( "Invalid counts array" );
		double[] volts = new double[ counts.length ];
		for( int index = 0; index < counts.length; index++ )
			volts[ index ] = countsToVolts( counts[ index ] );
		return volts;
	}	// countsToVolts()

	/**
	 * Converts a single voltage value to D/A counts, based on the current range setting.
	 * @param volts the voltage value to convert to counts.
	 * @return A count value calculated using the <i>current</i> D/A range. The count value returned is constrained
	 * to the minimum-maximum count range of the D/A.
	 * <i>(see {@link #setRange( int range ) setRange()})</i>.
	 */

	public char voltsToCounts( double volts ) {
		assert range >= DAC_RANGE_0_5V
			&& range <= DAC_RANGE_10V;
		int counts = ( int ) Math.round( MAX_COUNTS * ( volts - daRange[ range ][ 0 ] ) / daRange[ range ][ 1 ] );
		if( counts < 0 )
			counts = 0;
		else if( counts > MAX_COUNTS )
			counts = MAX_COUNTS;
		return ( char ) counts;
	}	// voltsToCounts()

	/**
	 * Converts an array of voltage values to an array of D/A count values, based on the current range setting.
	 * @param volts the array of voltage values to convert to counts.
	 * @return An array of count values calculated using the <i>current</i> D/A range. Each count value returned
	 * is constrained to the minimum-maximum count range of the D/A.
	 * <i>(see {@link #setRange( int range ) setRange()})</i>.
	 * @throws IllegalArgumentException
	 */

	public char[] voltsToCounts( double[] volts ) {
		if(
			volts == null
			|| volts.length < 1
		)
			throw new IllegalArgumentException( "Invalid volts array" );
		char[] counts = new char[ volts.length ];
		for( int index = 0; index < volts.length; index++ )
			counts[ index ] = voltsToCounts( volts[ index ] );
		return counts;
	}	// voltsToCounts()

	// }}}

}	// class AnalogOutputSubsystem

/* end of file */
