/*
 * $RCSfile: AIOUSB_ADC.c,v $
 * $Revision: 1.33 $
 * $Date: 2009/11/26 20:56:07 $
 * jEdit:tabSize=4:indentSize=4:collapseFolds=1:
 *
 * ACCES I/O USB API for Linux
 */


// {{{ includes
#include "AIOUSB_Core.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
// }}}

// {{{ C++ support
#ifdef __cplusplus
namespace AIOUSB {
#endif
// }}}

// {{{ local variables and functions

static const struct ADRange {
	double minVolts;
	double range;
} adRanges[ AD_NUM_GAIN_CODES ] = {
	  {   0, 10 }								// AD_GAIN_CODE_0_10V
	, { -10, 20 }								// AD_GAIN_CODE_10V
	, {   0,  5 }								// AD_GAIN_CODE_0_5V
	, {  -5, 10 }								// AD_GAIN_CODE_5V
	, {   0,  2 }								// AD_GAIN_CODE_0_2V
	, {  -2,  4 }								// AD_GAIN_CODE_2V
	, {   0,  1 }								// AD_GAIN_CODE_0_1V
	, {  -1,  2 }								// AD_GAIN_CODE_1V
};	// adRanges[]

// formerly public in the API
static unsigned long ADC_GetImmediate(
	unsigned long DeviceIndex
	, unsigned long Channel
	, unsigned short *pData );

// {{{ ReadConfigBlock()
static unsigned long ReadConfigBlock( unsigned long DeviceIndex, AIOUSB_BOOL forceRead ) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if(
		forceRead
		|| deviceDesc->cachedConfigBlock.size == 0
	) {
		libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
		if( deviceHandle != NULL ) {
			// request A/D configuration block from device
			ADConfigBlock configBlock;
			configBlock.device = deviceDesc;
			configBlock.size = deviceDesc->ConfigBytes;
			const unsigned timeout = deviceDesc->commTimeout;
			AIOUSB_UnLock();					// unlock while communicating with device
			const int bytesTransferred = libusb_control_transfer( deviceHandle
				, USB_READ_FROM_DEVICE, AUR_ADC_GET_CONFIG
				, 0, 0, configBlock.registers, configBlock.size, timeout );
			if( bytesTransferred == ( int ) configBlock.size ) {
				/*
				 * check and correct settings read from device
				 */
				AIOUSB_Lock();
				unsigned channel;
				for( channel = 0; channel < AD_NUM_GAIN_CODE_REGISTERS; channel++ ) {
					if(
						(
							configBlock.registers[ AD_CONFIG_GAIN_CODE + channel ]
							& ~( unsigned char ) ( AD_DIFFERENTIAL_MODE | AD_GAIN_CODE_MAX )
						) != 0
					)
						configBlock.registers[ AD_CONFIG_GAIN_CODE + channel ] = AD_GAIN_CODE_MIN;
				}	// for( channel ...

				const unsigned char calMode = configBlock.registers[ AD_CONFIG_CAL_MODE ];
				if(
					calMode != AD_CAL_MODE_NORMAL
					&& calMode != AD_CAL_MODE_GROUND
					&& calMode != AD_CAL_MODE_REFERENCE
				)
					configBlock.registers[ AD_CONFIG_CAL_MODE ] = AD_CAL_MODE_NORMAL;

				if( ( configBlock.registers[ AD_CONFIG_TRIG_COUNT ] & ~AD_TRIGGER_VALID_MASK ) != 0 )
					configBlock.registers[ AD_CONFIG_TRIG_COUNT ] = 0;

				const unsigned endChannel = AIOUSB_GetEndChannel( &configBlock );
				if(
					endChannel >= ( unsigned ) deviceDesc->ADCMUXChannels
					|| AIOUSB_GetStartChannel( &configBlock ) > endChannel
				)
					AIOUSB_SetScanRange( &configBlock, 0, deviceDesc->ADCMUXChannels - 1 );

				deviceDesc->cachedConfigBlock = configBlock;
				AIOUSB_UnLock();
			} else
				result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
		} else {
			result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
			AIOUSB_UnLock();
		}	// if( deviceHandle ...
	} else
		AIOUSB_UnLock();
	return result;
}	// ReadConfigBlock()
// }}}

// {{{ WriteConfigBlock()
static unsigned long WriteConfigBlock( unsigned long DeviceIndex ) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->cachedConfigBlock.size > 0 ) {
		libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
		if( deviceHandle != NULL ) {
			ADConfigBlock configBlock;
			configBlock = deviceDesc->cachedConfigBlock;
			const unsigned timeout = deviceDesc->commTimeout;
			AIOUSB_UnLock();					// unlock while communicating with device
			assert( configBlock.size > 0
				&& configBlock.size <= AD_MAX_CONFIG_REGISTERS );
			const int bytesTransferred = libusb_control_transfer( deviceHandle
				, USB_WRITE_TO_DEVICE, AUR_ADC_SET_CONFIG
				, 0, 0, configBlock.registers, configBlock.size, timeout );
			if( bytesTransferred != ( int ) configBlock.size )
				result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
		} else {
			result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
			AIOUSB_UnLock();
		}	// if( deviceHandle ...
	} else {
		result = AIOUSB_ERROR_INVALID_DATA;
		AIOUSB_UnLock();
	}	// if( deviceDesc->cachedConfigBlock.size ...

	return result;
}	// WriteConfigBlock()
// }}}

// }}}

// {{{ library private functions

// {{{ AIOUSB_GetScan()
PRIVATE unsigned long AIOUSB_GetScan(
	unsigned long DeviceIndex
	, unsigned short counts[]
) {
	if( counts == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
	if( deviceHandle != NULL ) {
		AIOUSB_UnLock();						// unlock while communicating with device
		result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
		AIOUSB_Lock();
		if( result == AIOUSB_SUCCESS ) {
			const ADConfigBlock origConfigBlock = deviceDesc->cachedConfigBlock;	// restore when done
			AIOUSB_BOOL configChanged = AIOUSB_FALSE
				, discardFirstSample = deviceDesc->discardFirstSample;
			unsigned startChannel = AIOUSB_GetStartChannel( &deviceDesc->cachedConfigBlock )
				, endChannel = AIOUSB_GetEndChannel( &deviceDesc->cachedConfigBlock )
				, overSample = AIOUSB_GetOversample( &deviceDesc->cachedConfigBlock );
			assert( startChannel <= endChannel );
			int numChannels = endChannel - startChannel + 1;

			/*
			 * in theory, all the A/D functions, including AIOUSB_GetScan(), should work in all
			 * measurement modes, including calibration mode; in practice, however, the device
			 * will return only a single sample in calibration mode; therefore, users must be
			 * careful to select a single channel and set oversample to zero during calibration
			 * mode; attempting to read more than one channel or use an oversample setting of
			 * more than zero in calibration mode will result in a timeout error; as a convenience
			 * to the user we automatically impose this restriction here in AIOUSB_GetScan(); if
			 * the device is changed to permit normal use of the A/D functions in calibration
			 * mode, we will have to modify this function to somehow recognize which devices
			 * support that capability, or simply delete this restriction altogether and rely on
			 * the users' good judgment
			 */
			const unsigned calMode = AIOUSB_GetCalMode( &deviceDesc->cachedConfigBlock );
			if(
				calMode == AD_CAL_MODE_GROUND
				|| calMode == AD_CAL_MODE_REFERENCE
			) {
				if( numChannels > 1 ) {
					AIOUSB_SetScanRange( &deviceDesc->cachedConfigBlock, startChannel, endChannel = startChannel );
					numChannels = 1;
					configChanged = AIOUSB_TRUE;
				}	// if( numChannels ...
				if( overSample > 0 ) {
					AIOUSB_SetOversample( &deviceDesc->cachedConfigBlock, overSample = 0 );
					configChanged = AIOUSB_TRUE;
				}	// if( overSample ...
				discardFirstSample = AIOUSB_FALSE;		// this feature can't be used in calibration mode either
			}	// if( calMode ...

			/*
			 * turn scan on and turn timer and external trigger off
			 */
			const unsigned origTriggerMode = AIOUSB_GetTriggerMode( &deviceDesc->cachedConfigBlock );
			unsigned triggerMode = origTriggerMode;
			triggerMode |= AD_TRIGGER_SCAN;								// enable scan
			triggerMode &= ~( AD_TRIGGER_TIMER | AD_TRIGGER_EXTERNAL );	// disable timer and external trigger
			if( triggerMode != origTriggerMode ) {
				AIOUSB_SetTriggerMode( &deviceDesc->cachedConfigBlock, triggerMode );
				configChanged = AIOUSB_TRUE;
			}	// if( triggerMode ...

			/*
			 * the oversample setting dictates how many samples to take _in addition_ to the primary
			 * sample; if oversample is zero, we take just one sample for each channel; if oversample
			 * is greater than zero then we average the primary sample and all of its over-samples; if
			 * the discardFirstSample setting is enabled, then we discard the primary sample, leaving
			 * just the over-samples; thus, if discardFirstSample is enabled, we must take at least one
			 * over-sample in order to have any data left; there's another complication: the device
			 * buffer is limited to a small number of samples, so we have to limit the number of
			 * over-samples to what the device buffer can accommodate, so the actual oversample setting
			 * depends on the number of channels being scanned; we also preserve and restore the original
			 * oversample setting specified by the user; since the user is expecting to average
			 * (1 + oversample) samples, then if discardFirstSample is enabled we simply always add one
			 */
			const unsigned origOverSample = overSample;
			int samplesPerChannel = 1 + origOverSample;
			if( discardFirstSample )
				samplesPerChannel++;
			if( samplesPerChannel > 256 )
				samplesPerChannel = 256;		// constrained by maximum oversample of 255

			/*
			 * make sure device buffer can accommodate this number of samples
			 */
			const int DEVICE_SAMPLE_BUFFER_SIZE = 1024;		// number of samples device can buffer
			if( ( numChannels * samplesPerChannel ) > DEVICE_SAMPLE_BUFFER_SIZE )
				samplesPerChannel = DEVICE_SAMPLE_BUFFER_SIZE / numChannels;
			overSample = samplesPerChannel - 1;
			if( overSample != origOverSample ) {
				AIOUSB_SetOversample( &deviceDesc->cachedConfigBlock, overSample );
				configChanged = AIOUSB_TRUE;
			}	// if( overSample ...

			if( configChanged ) {
				AIOUSB_UnLock();				// unlock while communicating with device
				result = WriteConfigBlock( DeviceIndex );
				AIOUSB_Lock();
			}	// if( configChanged )
			if( result == AIOUSB_SUCCESS ) {
				const int numSamples = numChannels * samplesPerChannel;
				const unsigned short numSamplesHigh = ( unsigned short ) ( numSamples >> 16 );
				const unsigned short numSamplesLow = ( unsigned short ) numSamples;
				const int numBytes = numSamples * sizeof( unsigned short );
				unsigned short *const sampleBuffer = ( unsigned short * ) malloc( numBytes );
				assert( sampleBuffer != 0 );
				if( sampleBuffer != 0 ) {
					const unsigned timeout = deviceDesc->commTimeout;
					AIOUSB_UnLock();			// unlock while communicating with device
					int bytesTransferred = libusb_control_transfer( deviceHandle
						, USB_WRITE_TO_DEVICE, AUR_START_ACQUIRING_BLOCK
						, numSamplesHigh, numSamplesLow, 0, 0, timeout );
					if( bytesTransferred == 0 ) {
						// request AUR_ADC_IMMEDIATE triggers the sampling of data
						bytesTransferred = libusb_control_transfer( deviceHandle
							, USB_READ_FROM_DEVICE, AUR_ADC_IMMEDIATE
							, 0, 0, ( unsigned char * ) sampleBuffer /* discarded */, sizeof( unsigned short )
							, timeout );
						if( bytesTransferred == sizeof( unsigned short ) ) {
							const int libusbResult = AIOUSB_BulkTransfer( deviceHandle
								, LIBUSB_ENDPOINT_IN | USB_BULK_READ_ENDPOINT
								, ( unsigned char * ) sampleBuffer, numBytes, &bytesTransferred
								, timeout );
							if( libusbResult != LIBUSB_SUCCESS ) {
								result = LIBUSB_RESULT_TO_AIOUSB_RESULT( libusbResult );
							} else if( bytesTransferred != numBytes ) {
								result = AIOUSB_ERROR_INVALID_DATA;
							} else {
								/*
								 * compute the average of all the samples taken for each channel, discarding
								 * the first sample if that option is enabled; each byte in sampleBuffer[] is
								 * 1 of 2 bytes for each sample, the first byte being the LSB and the second
								 * byte the MSB, in other words, little-endian format; so for convenience we
								 * simply declare sampleBuffer[] to be of type 'unsigned short' and the data
								 * is already in the correct format; the device returns data only for the
								 * channels requested, from startChannel to endChannel; AIOUSB_GetScan()
								 * returns the averaged data readings in counts[], putting the reading for
								 * startChannel in counts[0], and the reading for endChannel in counts[numChannels-1]
								 */
								const int samplesToAverage
									= discardFirstSample
										? samplesPerChannel - 1
										: samplesPerChannel;
								int sampleIndex = 0;
								int channel;
								for( channel = 0; channel < numChannels; channel++ ) {
									unsigned long sampleSum = 0;
									if( discardFirstSample )
										sampleIndex++;		// skip over first sample
									int sample;
									for( sample = 0; sample < samplesToAverage; sample++ )
										sampleSum += sampleBuffer[ sampleIndex++ ];
									counts[ channel ] = ( unsigned short )
										( ( sampleSum + samplesToAverage / 2 ) / samplesToAverage );
								}	// for( channel ...
							}	// else if( bytesTransferred ...
						} else
							result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
					} else
						result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
					free( sampleBuffer );
#if defined( POST_ACQUIRE_RESET )
					AIOUSB_Reset( DeviceIndex );
#endif
				} else {
					result = AIOUSB_ERROR_NOT_ENOUGH_MEMORY;
					AIOUSB_UnLock();
				}	// if( sampleBuffer ...
			} else
				AIOUSB_UnLock();
			if( configChanged ) {
				AIOUSB_Lock();
				deviceDesc->cachedConfigBlock = origConfigBlock;
				AIOUSB_UnLock();				// unlock while communicating with device
				WriteConfigBlock( DeviceIndex );
			}	// if( configChanged )
		} else
			AIOUSB_UnLock();
	} else {
		result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
		AIOUSB_UnLock();
	}	// if( deviceHandle ...

	return result;
}	// AIOUSB_GetScan()
// }}}

// {{{ AIOUSB_ArrayCountsToVolts()
PRIVATE unsigned long AIOUSB_ArrayCountsToVolts(
	unsigned long DeviceIndex
	, int startChannel
	, int numChannels
	, const unsigned short counts[]
	, double volts[]
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	assert( startChannel >= 0
		&& numChannels >= 0
		&& startChannel + numChannels <= ( int ) deviceDesc->ADCMUXChannels
		&& counts != 0
		&& volts != 0 );
	if(
		startChannel < 0
		|| numChannels < 0
		|| startChannel + numChannels > ( int ) deviceDesc->ADCMUXChannels
		|| counts == NULL
		|| volts == NULL
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( endChannel ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		int channel;
		for( channel = 0; channel < numChannels; channel++ ) {
			const int gainCode = AIOUSB_GetGainCode( &deviceDesc->cachedConfigBlock, startChannel + channel );
			assert( gainCode >= AD_GAIN_CODE_MIN
				&& gainCode <= AD_GAIN_CODE_MAX );
			const struct ADRange *const range = &adRanges[ gainCode ];
			volts[ channel ]
				= (
					( ( double ) counts[ channel ] / ( double ) AI_16_MAX_COUNTS )
					* range->range
				)
				+ range->minVolts;
		}	// for( channel ...
		AIOUSB_UnLock();
	}	// if( result ...

	return result;
}	// AIOUSB_ArrayCountsToVolts()
// }}}

// {{{ AIOUSB_ArrayVoltsToCounts()
PRIVATE unsigned long AIOUSB_ArrayVoltsToCounts(
	unsigned long DeviceIndex
	, int startChannel
	, int numChannels
	, const double volts[]
	, unsigned short counts[]
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	assert( startChannel >= 0
		&& numChannels >= 0
		&& startChannel + numChannels <= ( int ) deviceDesc->ADCMUXChannels
		&& volts != 0
		&& counts != 0 );
	if(
		startChannel < 0
		|| numChannels < 0
		|| startChannel + numChannels > ( int ) deviceDesc->ADCMUXChannels
		|| volts == NULL
		|| counts == NULL
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( endChannel ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		int channel;
		for( channel = 0; channel < numChannels; channel++ ) {
			const int gainCode = AIOUSB_GetGainCode( &deviceDesc->cachedConfigBlock, startChannel + channel );
			assert( gainCode >= AD_GAIN_CODE_MIN
				&& gainCode <= AD_GAIN_CODE_MAX );
			const struct ADRange *const range = &adRanges[ gainCode ];
			int rawCounts = round(
				( double ) AI_16_MAX_COUNTS
				* ( volts[ channel ] - range->minVolts )
				/ range->range
			);
			if( rawCounts < 0 )
				rawCounts = 0;
			else if( rawCounts > AI_16_MAX_COUNTS )
				rawCounts = AI_16_MAX_COUNTS;
			counts[ channel ] = ( unsigned short ) rawCounts;
		}	// for( channel ...
		AIOUSB_UnLock();
	}	// if( result ...

	return result;
}	// AIOUSB_ArrayVoltsToCounts()
// }}}

// }}}

// {{{ ADC_GetChannelV()
unsigned long ADC_GetChannelV(
	unsigned long DeviceIndex
	, unsigned long ChannelIndex
	, double *pBuf
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...
	if(
		pBuf == NULL
		|| ChannelIndex >= deviceDesc->ADCMUXChannels
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( pBuf ...

	/*
	 * there is no guarantee that ChannelIndex, passed by the user, is within the current
	 * channel scan range; if it is not, then valid data cannot be returned; in addition,
	 * since we're only returning the data for a single channel, there's no need to scan
	 * all the channels; the Windows implementation attempts to improve performance by caching
	 * all the values read; but the technique is riddled with problems; first of all, it can
	 * easily return extremely stale data, without any indication to the user; secondly, it
	 * can return data for channels that weren't even scanned, without any indication to the
	 * user; thirdly, caching is unnecessary; if the user wants to read a single channel they
	 * can call ADC_GetChannelV(); if the user wants to improve performance by reading multiple
	 * channels they can call ADC_GetScanV(); so to address all these issues, we temporarily
	 * compress the scan range to just ChannelIndex and then restore it when we're done; so
	 * in this implementation all calls to ADC_GetChannelV() return "real-time" data for the
	 * specified channel
	 */
	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		const ADConfigBlock origConfigBlock = deviceDesc->cachedConfigBlock;	// restore when done
		AIOUSB_SetScanRange( &deviceDesc->cachedConfigBlock, ChannelIndex, ChannelIndex );
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
		unsigned short counts;
		result = AIOUSB_GetScan( DeviceIndex, &counts );
		if( result == AIOUSB_SUCCESS ) {
			double volts;
			result = AIOUSB_ArrayCountsToVolts( DeviceIndex, ChannelIndex, 1, &counts, &volts );
			if( result == AIOUSB_SUCCESS )
				*pBuf = volts;
			else
				*pBuf = 0.0;
		}	// if( result ...
		AIOUSB_Lock();
		deviceDesc->cachedConfigBlock = origConfigBlock;
		AIOUSB_UnLock();						// unlock while communicating with device
		WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_GetChannelV()
// }}}

// {{{ ADC_GetScanV()
unsigned long ADC_GetScanV(
	unsigned long DeviceIndex
	, double *pBuf
) {
	if( pBuf == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result == AIOUSB_SUCCESS ) {
		DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
		if( deviceDesc->bADCStream ) {
			/*
			 * get raw A/D counts
			 */
			unsigned short *const counts
				= ( unsigned short * ) malloc( deviceDesc->ADCMUXChannels * sizeof( unsigned short ) );
			assert( counts != 0 );
			if( counts != 0 ) {
				AIOUSB_UnLock();				// unlock while communicating with device
				result = ADC_GetScan( DeviceIndex, counts );
				AIOUSB_Lock();
				if( result == AIOUSB_SUCCESS ) {
					/*
					 * convert from A/D counts to volts; only the channels from startChannel to
					 * endChannel contain valid data, so we only convert those; pBuf[] is
					 * expected to contain entries for all the A/D channels; so for cleanliness,
					 * we zero out the channels in pBuf[] that aren't going to be filled in with
					 * real readings
					 */
					const unsigned startChannel = AIOUSB_GetStartChannel( &deviceDesc->cachedConfigBlock )
						, endChannel = AIOUSB_GetEndChannel( &deviceDesc->cachedConfigBlock );
					assert( startChannel <= endChannel );

					/*
					 * zero out unused channels
					 */
					unsigned channel;
					for( channel = 0; channel < deviceDesc->ADCMUXChannels; channel++ ) {
						if(
							channel < startChannel
							|| channel > endChannel
						)
							pBuf[ channel ] = 0.0;
					}	// for( channel ...

					/*
					 * convert remaining channels to volts
					 */
					result = AIOUSB_ArrayCountsToVolts( DeviceIndex, startChannel, endChannel - startChannel + 1
						, counts + startChannel, pBuf + startChannel );
				}	// if( result ...
				free( counts );
			} else
				result = AIOUSB_ERROR_NOT_ENOUGH_MEMORY;
		} else
			result = AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( result ...

	AIOUSB_UnLock();
	return result;
}	// ADC_GetScanV()
// }}}

// {{{ ADC_GetScan()
unsigned long ADC_GetScan(
	unsigned long DeviceIndex
	, unsigned short *pBuf
) {
	if( pBuf == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	/*
	 * pBuf[] is expected to contain entries for all the A/D channels, even though we may
	 * be reading only a few channels; so for cleanliness, we zero out the channels in
	 * pBuf[] that aren't going to be filled in with real readings
	 */
	memset( pBuf, 0, deviceDesc->ADCMUXChannels * sizeof( unsigned short ) );
	const unsigned startChannel = AIOUSB_GetStartChannel( &deviceDesc->cachedConfigBlock );
	AIOUSB_UnLock();
	return AIOUSB_GetScan( DeviceIndex, pBuf + startChannel );
}	// ADC_GetScan()
// }}}

// {{{ ADC_GetConfig()
unsigned long ADC_GetConfig(
	unsigned long DeviceIndex
	, unsigned char *pConfigBuf
	, unsigned long *ConfigBufSize
) {
	if(
		pConfigBuf == NULL
		|| ConfigBufSize == NULL
	)
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->ConfigBytes == 0 ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->ConfigBytes ...
	if( *ConfigBufSize < deviceDesc->ConfigBytes ) {
		*ConfigBufSize = deviceDesc->ConfigBytes;
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( *ConfigBufSize ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_TRUE );
	if( result == AIOUSB_SUCCESS ) {
		assert( deviceDesc->cachedConfigBlock.size > 0
			&& deviceDesc->cachedConfigBlock.size <= AD_MAX_CONFIG_REGISTERS );
		AIOUSB_Lock();
		memcpy( pConfigBuf, deviceDesc->cachedConfigBlock.registers, deviceDesc->cachedConfigBlock.size );
		*ConfigBufSize = deviceDesc->cachedConfigBlock.size;
		AIOUSB_UnLock();
	}	// if( result ...

	return result;
}	// ADC_GetConfig()
// }}}

// {{{ ADC_SetConfig()
unsigned long ADC_SetConfig(
	unsigned long DeviceIndex
	, unsigned char *pConfigBuf
	, unsigned long *ConfigBufSize
) {
	if(
		pConfigBuf == NULL
		|| ConfigBufSize == NULL
	)
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->ConfigBytes == 0 ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->ConfigBytes ...
	if( *ConfigBufSize < deviceDesc->ConfigBytes ) {
		*ConfigBufSize = deviceDesc->ConfigBytes;
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( *ConfigBufSize ...

	/*
	 * validate settings
	 */
	ADConfigBlock configBlock;
	configBlock.device = deviceDesc;
	configBlock.size = deviceDesc->ConfigBytes;
	assert( configBlock.size > 0
		&& configBlock.size <= AD_MAX_CONFIG_REGISTERS );
	memcpy( configBlock.registers, pConfigBuf, configBlock.size );

	unsigned channel;
	for( channel = 0; channel < AD_NUM_GAIN_CODE_REGISTERS; channel++ ) {
		if(
			(
				configBlock.registers[ AD_CONFIG_GAIN_CODE + channel ]
				& ~( unsigned char ) ( AD_DIFFERENTIAL_MODE | AD_GAIN_CODE_MAX )
			) != 0
		) {
			AIOUSB_UnLock();
			return AIOUSB_ERROR_INVALID_PARAMETER;
		}	// if( ( configBlock.registers[ ...
	}	// for( channel ...

	const unsigned char calMode = configBlock.registers[ AD_CONFIG_CAL_MODE ];
	if(
		calMode != AD_CAL_MODE_NORMAL
		&& calMode != AD_CAL_MODE_GROUND
		&& calMode != AD_CAL_MODE_REFERENCE
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( calMode ...

	if( ( configBlock.registers[ AD_CONFIG_TRIG_COUNT ] & ~AD_TRIGGER_VALID_MASK ) != 0 ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( ( configBlock.registers[ ...

	const unsigned endChannel = AIOUSB_GetEndChannel( &configBlock );
	if(
		endChannel >= ( unsigned ) deviceDesc->ADCMUXChannels
		|| AIOUSB_GetStartChannel( &configBlock ) > endChannel
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( endChannel ...

	deviceDesc->cachedConfigBlock = configBlock;
	AIOUSB_UnLock();							// unlock while communicating with device
	result = WriteConfigBlock( DeviceIndex );
	if( result == AIOUSB_SUCCESS )
		*ConfigBufSize = configBlock.size;

	return result;
}	// ADC_SetConfig()
// }}}

// {{{ ADC_RangeAll()
unsigned long ADC_RangeAll(
	unsigned long DeviceIndex
	, unsigned char *pGainCodes
	, unsigned long bSingleEnded
) {
	if(
		pGainCodes == NULL
		|| (
			bSingleEnded != AIOUSB_FALSE
			&& bSingleEnded != AIOUSB_TRUE
		)
	)
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if(
		deviceDesc->ADCChannels == 0
		|| deviceDesc->bADCStream == AIOUSB_FALSE
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->ADCChannels ...

	/*
	 * validate gain codes; they should be just gain codes; single-ended or differential
	 * mode is specified by bSingleEnded
	 */
	unsigned channel;
	for( channel = 0; channel < deviceDesc->ADCChannels; channel++ ) {
		if( ( pGainCodes[ AD_CONFIG_GAIN_CODE + channel ] & ~AD_GAIN_CODE_MAX ) != 0 ) {
			AIOUSB_UnLock();
			return AIOUSB_ERROR_INVALID_PARAMETER;
		}	// if( ( pGainCodes[ ...
	}	// for( channel ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		for( channel = 0; channel < deviceDesc->ADCChannels; channel++ ) {
			AIOUSB_SetGainCode( &deviceDesc->cachedConfigBlock, channel, pGainCodes[ channel ] );
			AIOUSB_SetDifferentialMode( &deviceDesc->cachedConfigBlock, channel
				, ( bSingleEnded == AIOUSB_FALSE ) ? AIOUSB_TRUE : AIOUSB_FALSE );
		}	// for( channel ...
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_RangeAll()
// }}}

// {{{ ADC_Range1()
unsigned long ADC_Range1(
	unsigned long DeviceIndex
	, unsigned long ADChannel
	, unsigned char GainCode
	, unsigned long bSingleEnded
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if(
		deviceDesc->ADCMUXChannels == 0
		|| deviceDesc->bADCStream == AIOUSB_FALSE
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->ADCMUXChannels ...

	if(
		( GainCode & ~AD_GAIN_CODE_MAX ) != 0
		|| (
			bSingleEnded != AIOUSB_FALSE
			&& bSingleEnded != AIOUSB_TRUE
		)
		|| ADChannel >= deviceDesc->ADCMUXChannels
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( ( GainCode ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		AIOUSB_SetGainCode( &deviceDesc->cachedConfigBlock, ADChannel, GainCode );
		AIOUSB_SetDifferentialMode( &deviceDesc->cachedConfigBlock, ADChannel
			, ( bSingleEnded == AIOUSB_FALSE ) ? AIOUSB_TRUE : AIOUSB_FALSE );
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_Range1()
// }}}

// {{{ ADC_ADMode()
unsigned long ADC_ADMode(
	unsigned long DeviceIndex
	, unsigned char TriggerMode
	, unsigned char CalMode
) {
	if(
		( TriggerMode & ~AD_TRIGGER_VALID_MASK ) != 0
		|| (
			CalMode != AD_CAL_MODE_NORMAL
			&& CalMode != AD_CAL_MODE_GROUND
			&& CalMode != AD_CAL_MODE_REFERENCE
		)
	)
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		AIOUSB_SetCalMode( &deviceDesc->cachedConfigBlock, CalMode );
		AIOUSB_SetTriggerMode( &deviceDesc->cachedConfigBlock, TriggerMode );
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_ADMode()
// }}}

// {{{ ADC_SetOversample()
unsigned long ADC_SetOversample(
	unsigned long DeviceIndex
	, unsigned char Oversample
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		AIOUSB_SetOversample( &deviceDesc->cachedConfigBlock, Oversample );
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_SetOversample()
// }}}

// {{{ ADC_SetScanLimits()
unsigned long ADC_SetScanLimits(
	unsigned long DeviceIndex
	, unsigned long StartChannel
	, unsigned long EndChannel
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	if(
		EndChannel > deviceDesc->ADCMUXChannels
		|| StartChannel > EndChannel
	) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_INVALID_PARAMETER;
	}	// if( EndChannel ...

	AIOUSB_UnLock();							// unlock while communicating with device
	result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
	if( result == AIOUSB_SUCCESS ) {
		AIOUSB_Lock();
		AIOUSB_SetScanRange( &deviceDesc->cachedConfigBlock, StartChannel, EndChannel );
		AIOUSB_UnLock();						// unlock while communicating with device
		result = WriteConfigBlock( DeviceIndex );
	}	// if( result ...

	return result;
}	// ADC_SetScanLimits()
// }}}

// {{{ ADC_SetCal()
unsigned long ADC_SetCal(
	unsigned long DeviceIndex
	, const char *CalFileName
) {
	if( CalFileName == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	assert( strlen( CalFileName ) > 0 );
	unsigned long result;
	if( strcmp( CalFileName, ":AUTO:" ) == 0 )
		result = AIOUSB_ADC_InternalCal( DeviceIndex, AIOUSB_TRUE, 0, 0 );
	else if(
		strcmp( CalFileName, ":NONE:" ) == 0
		|| strcmp( CalFileName, ":1TO1:" ) == 0
	)
		result = AIOUSB_ADC_InternalCal( DeviceIndex, AIOUSB_FALSE, 0, 0 );
	else
		result = AIOUSB_ADC_LoadCalTable( DeviceIndex, CalFileName );

	return result;
}	// ADC_SetCal()
// }}}

// {{{ ADC_QueryCal()
unsigned long ADC_QueryCal(
	unsigned long DeviceIndex
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
	if( deviceHandle != NULL ) {
		const unsigned timeout = deviceDesc->commTimeout;
		AIOUSB_UnLock();						// unlock while communicating with device
		unsigned char calSupported = 0xff;		// so we can detect if it changes
		const int bytesTransferred = libusb_control_transfer( deviceHandle, USB_READ_FROM_DEVICE, AUR_PROBE_CALFEATURE
			, 0, 0, &calSupported, sizeof( calSupported ), timeout );
		if( bytesTransferred == sizeof( calSupported ) ) {
			if( calSupported != 0xBB )			// 0xBB == AUR_LOAD_BULK_CALIBRATION_BLOCK
				result = AIOUSB_ERROR_NOT_SUPPORTED;
		} else
			result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
	} else {
		result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
		AIOUSB_UnLock();
	}	// if( deviceHandle ...

	return result;
}	// ADC_QueryCal()
// }}}

// {{{ ADC_Initialize()
unsigned long ADC_Initialize(
	unsigned long DeviceIndex
	, unsigned char *pConfigBuf
	, unsigned long *ConfigBufSize
	, const char *CalFileName
) {
	unsigned long result = AIOUSB_SUCCESS;

	if(
		pConfigBuf != NULL
		&& ConfigBufSize != NULL
	)
		result = ADC_SetConfig( DeviceIndex, pConfigBuf, ConfigBufSize );

	if(
		result == AIOUSB_SUCCESS
		&& CalFileName != NULL
	)
		result = ADC_SetCal( DeviceIndex, CalFileName );

	return result;
}	// ADC_Initialize()
// }}}

// {{{ ADC_BulkAcquire()

static void *BulkAcquireWorker( void *params );

unsigned long ADC_BulkAcquire(
	unsigned long DeviceIndex
	, unsigned long BufSize
	, void *pBuf
) {
	if( pBuf == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	if( deviceDesc->workerBusy ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_OPEN_FAILED;
	}	// if( deviceDesc->workerBusy )

	AIOUSB_UnLock();
	struct BulkAcquireWorkerParams *const acquireParams
		= ( struct BulkAcquireWorkerParams * ) malloc( sizeof( struct BulkAcquireWorkerParams ) );
	assert( acquireParams != 0 );
	if( acquireParams != 0 ) {
		acquireParams->DeviceIndex = DeviceIndex;
		acquireParams->BufSize = BufSize;
		acquireParams->pBuf = pBuf;
		const int maxPriority = sched_get_priority_max( SCHED_FIFO );
		struct sched_param schedParam = { maxPriority };
		pthread_attr_t workerThreadAttr;
		pthread_t workerThreadID;
		pthread_attr_init( &workerThreadAttr );
		pthread_attr_setschedpolicy( &workerThreadAttr, SCHED_FIFO );
		pthread_attr_setschedparam( &workerThreadAttr, &schedParam );
		const int threadResult = pthread_create( &workerThreadID, &workerThreadAttr, BulkAcquireWorker, acquireParams );
		if( threadResult == 0 ) {
			sched_yield();
		} else {
			free( acquireParams );
			result = AIOUSB_ERROR_INVALID_THREAD;
		}	// if( threadResult ...
		pthread_attr_destroy( &workerThreadAttr );
	} else
		result = AIOUSB_ERROR_NOT_ENOUGH_MEMORY;

	return result;
}	// ADC_BulkAcquire()


static void *BulkAcquireWorker( void *params ) {
	/*
	 * we assume the parameters passed to BulkAcquireWorker() have
	 * been validated by ADC_BulkAcquire()
	 */
	assert( params != 0 );
	unsigned long result = AIOUSB_SUCCESS;
	struct BulkAcquireWorkerParams *const acquireParams = ( struct BulkAcquireWorkerParams * ) params;
	AIOUSB_Lock();
	DeviceDescriptor *const deviceDesc = &deviceTable[ acquireParams->DeviceIndex ];
	libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( acquireParams->DeviceIndex );
	if( deviceHandle != NULL ) {
		unsigned long bytesRemaining = acquireParams->BufSize;
		deviceDesc->workerStatus = bytesRemaining;	// deviceDesc->workerStatus == bytes remaining to receive
		deviceDesc->workerResult = AIOUSB_SUCCESS;
		deviceDesc->workerBusy = AIOUSB_TRUE;
		double clockHz = deviceDesc->miscClockHz;
		const unsigned long streamingBlockSize = deviceDesc->StreamingBlockSize * sizeof( unsigned short );	// bytes
		const unsigned timeout = deviceDesc->commTimeout;
		AIOUSB_UnLock();						// unlock while communicating with device
		const unsigned short numSamplesHigh = ( unsigned short ) ( acquireParams->BufSize >> 17 );	// acquireParams->BufSize is bytes
		const unsigned short numSamplesLow = ( unsigned short ) ( acquireParams->BufSize >> 1 );
		unsigned char *data = ( unsigned char * ) acquireParams->pBuf;
		assert( data != 0 );

		int bytesTransferred = libusb_control_transfer( deviceHandle
			, USB_WRITE_TO_DEVICE, AUR_START_ACQUIRING_BLOCK
			, numSamplesHigh, numSamplesLow, 0, 0, timeout );
		if( bytesTransferred == 0 ) {
			CTR_StartOutputFreq( acquireParams->DeviceIndex, 0, &clockHz );
			while( bytesRemaining > 0 ) {
				unsigned long bytesToTransfer
					= ( bytesRemaining < streamingBlockSize )
						? bytesRemaining
						: streamingBlockSize;
				const int libusbResult = AIOUSB_BulkTransfer( deviceHandle
					, LIBUSB_ENDPOINT_IN | USB_BULK_READ_ENDPOINT
					, data, ( int ) bytesToTransfer, &bytesTransferred
					, timeout );
				if( libusbResult != LIBUSB_SUCCESS ) {
					result = LIBUSB_RESULT_TO_AIOUSB_RESULT( libusbResult );
					break;						// from while()
				} else if( bytesTransferred != ( int ) bytesToTransfer ) {
					result = AIOUSB_ERROR_INVALID_DATA;
					break;						// from while()
				} else {
					data += bytesTransferred;
					bytesRemaining -= bytesTransferred;
					AIOUSB_Lock();
					deviceDesc->workerStatus = bytesRemaining;
					AIOUSB_UnLock();
				}	// else if( bytesTransferred ...
			}	// while( bytesRemaining ...
			clockHz = 0;
			CTR_StartOutputFreq( acquireParams->DeviceIndex, 0, &clockHz );
		} else
			result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
#if defined( POST_ACQUIRE_RESET )
		AIOUSB_Reset( acquireParams->DeviceIndex );
#endif
	} else {
		result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
		AIOUSB_UnLock();
	}	// if( deviceHandle ...

	AIOUSB_Lock();
	deviceDesc->workerBusy = AIOUSB_FALSE;
	deviceDesc->workerStatus = 0;
	deviceDesc->workerResult = result;
	AIOUSB_UnLock();
	free( params );
	return 0;
}	// BulkAcquireWorker()

// }}}

// {{{ ADC_BulkPoll()
unsigned long ADC_BulkPoll(
	unsigned long DeviceIndex
	, unsigned long *BytesLeft
) {
	if( BytesLeft == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	*BytesLeft = deviceDesc->workerStatus;
	result = deviceDesc->workerResult;
	AIOUSB_UnLock();

	return result;
}	// ADC_BulkPoll()
// }}}

// {{{ ADC_GetImmediate()
/*
 * o this function is erroneously documented in the API specification, but it should
 *   not be made public; it is useful internally, however, for such things as
 *   calibrating the A/D converter
 *
 * o the specification does not include a Channel parameter, but this implementation
 *   does because the Pascal code does and because it's used by ADC_SetCal()
 *
 * o in a departure from the Pascal code, this function supports two categories of
 *   "immediate" A/Ds: the older products which have a single immediate A/D channel,
 *   and the newer products which have multiple immediate A/D channels; fortunately,
 *   this function accepts a pData parameter, which permits us to return any amount of
 *   data; the caller simply has to make sure that his pData buffer is large enough for
 *   the particular device; that's not an unreasonable demand since this function is
 *   used internally and not intended to be public
 */
static unsigned long ADC_GetImmediate(
	unsigned long DeviceIndex
	, unsigned long Channel
	, unsigned short *pData
) {
	if( pData == NULL )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->ImmADCs == 0 ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->ImmADCs ...

	libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
	if( deviceHandle != NULL ) {
		const unsigned timeout = deviceDesc->commTimeout;
		const int numBytes = sizeof( unsigned short ) * deviceDesc->ImmADCs;
		AIOUSB_UnLock();						// unlock while communicating with device
		const int bytesTransferred = libusb_control_transfer( deviceHandle, USB_READ_FROM_DEVICE, AUR_ADC_IMMEDIATE
			, 0, Channel, ( unsigned char * ) pData, numBytes, timeout );
		if( bytesTransferred != numBytes )
			result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
	} else {
		result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
		AIOUSB_UnLock();
	}	// if( deviceHandle ...

	return result;
}	// ADC_GetImmediate()
// }}}

// {{{ extended API functions

// {{{ AIOUSB_IsDiscardFirstSample()
AIOUSB_BOOL AIOUSB_IsDiscardFirstSample(
	unsigned long DeviceIndex
) {
	AIOUSB_BOOL discard = AIOUSB_FALSE;

	if( ! AIOUSB_Lock() )
		return discard;

	if( AIOUSB_Validate( &DeviceIndex ) == AIOUSB_SUCCESS )
		discard = deviceTable[ DeviceIndex ].discardFirstSample;

	AIOUSB_UnLock();
	return discard;
}	// AIOUSB_IsDiscardFirstSample()
// }}}

// {{{ AIOUSB_SetDiscardFirstSample()
unsigned long AIOUSB_SetDiscardFirstSample(
	unsigned long DeviceIndex
	, AIOUSB_BOOL discard
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result == AIOUSB_SUCCESS )
		deviceTable[ DeviceIndex ].discardFirstSample = discard;

	AIOUSB_UnLock();
	return result;
}	// AIOUSB_SetDiscardFirstSample()
// }}}

// {{{ AIOUSB_CountsToVolts()
double AIOUSB_CountsToVolts(
	unsigned long DeviceIndex
	, unsigned channel
	, unsigned short counts
) {
	double volts;
	if( AIOUSB_ArrayCountsToVolts( DeviceIndex, channel, 1, &counts, &volts ) != AIOUSB_SUCCESS )
		volts = 0.0;
	return volts;
}	// AIOUSB_CountsToVolts()
// }}}

// {{{ AIOUSB_MultipleCountsToVolts()
unsigned long AIOUSB_MultipleCountsToVolts(
	unsigned long DeviceIndex
	, unsigned startChannel
	, unsigned endChannel
	, const unsigned short counts[ /* deviceDesc->ADCMUXChannels */ ]
	, double volts[ /* deviceDesc->ADCMUXChannels */ ]
) {
	return AIOUSB_ArrayCountsToVolts( DeviceIndex, startChannel, endChannel - startChannel + 1
		, counts + startChannel, volts + startChannel );
}	// AIOUSB_MultipleCountsToVolts()
// }}}

// {{{ AIOUSB_VoltsToCounts()
unsigned short AIOUSB_VoltsToCounts(
	unsigned long DeviceIndex
	, unsigned channel
	, double volts
) {
	unsigned short counts;
	if( AIOUSB_ArrayVoltsToCounts( DeviceIndex, channel, 1, &volts, &counts ) != AIOUSB_SUCCESS )
		counts = 0;
	return counts;
}	// AIOUSB_VoltsToCounts()
// }}}

// {{{ AIOUSB_MultipleVoltsToCounts()
unsigned long AIOUSB_MultipleVoltsToCounts(
	unsigned long DeviceIndex
	, unsigned startChannel
	, unsigned endChannel
	, const double volts[ /* deviceDesc->ADCMUXChannels */ ]
	, unsigned short counts[ /* deviceDesc->ADCMUXChannels */ ]
) {
	return AIOUSB_ArrayVoltsToCounts( DeviceIndex, startChannel, endChannel - startChannel + 1
		, volts + startChannel, counts + startChannel );
}	// AIOUSB_MultipleVoltsToCounts()
// }}}

// {{{ AIOUSB_ADC_LoadCalTable()
unsigned long AIOUSB_ADC_LoadCalTable(
	unsigned long DeviceIndex
	, const char *fileName
) {
	if( fileName == 0 )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	if( ( result = ADC_QueryCal( DeviceIndex ) ) != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( ( result ...

	unsigned short *const calTable = ( unsigned short * ) malloc( CAL_TABLE_WORDS * sizeof( unsigned short ) );
	assert( calTable != 0 );
	if( calTable != 0 ) {
		FILE *const calFile = fopen( fileName, "r" );
		if( calFile != NULL ) {
			const size_t wordsRead = fread( calTable, sizeof( unsigned short ), CAL_TABLE_WORDS, calFile );
			fclose( calFile );
			if( wordsRead == ( size_t ) CAL_TABLE_WORDS )
				result = AIOUSB_ADC_SetCalTable( DeviceIndex, calTable );
			else
				result = AIOUSB_ERROR_FILE_NOT_FOUND;
		} else
			result = AIOUSB_ERROR_FILE_NOT_FOUND;
		free( calTable );
	} else
		result = AIOUSB_ERROR_NOT_ENOUGH_MEMORY;

	return result;
}	// AIOUSB_ADC_LoadCalTable()
// }}}

// {{{ AIOUSB_ADC_SetCalTable()
unsigned long AIOUSB_ADC_SetCalTable(
	unsigned long DeviceIndex
	, const unsigned short calTable[]
) {
	if( calTable == 0 )
		return AIOUSB_ERROR_INVALID_PARAMETER;

	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	if( ( result = ADC_QueryCal( DeviceIndex ) ) != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( ( result ...

	libusb_device_handle *const deviceHandle = AIOUSB_GetDeviceHandle( DeviceIndex );
	if( deviceHandle != NULL ) {
		/*
		 * send calibration table to SRAM one block at a time; according to the documentation,
		 * the proper procedure is to bulk transfer a block of calibration data to "endpoint 2"
		 * and then send a control message to load it into the SRAM
		 */
		const unsigned timeout = deviceDesc->commTimeout;
		AIOUSB_UnLock();						// unlock while communicating with device
		const int SRAM_BLOCK_WORDS = 1024;		// can send 1024 words at a time to SRAM
		int sramAddress = 0;
		int wordsRemaining = CAL_TABLE_WORDS;
		while( wordsRemaining > 0 ) {
			const int wordsWritten
				= ( wordsRemaining < SRAM_BLOCK_WORDS )
					? wordsRemaining
					: SRAM_BLOCK_WORDS;
			int bytesTransferred;
			const int libusbResult = AIOUSB_BulkTransfer( deviceHandle
				, LIBUSB_ENDPOINT_OUT | USB_BULK_WRITE_ENDPOINT
				, ( unsigned char * ) ( calTable + sramAddress ), wordsWritten * sizeof( unsigned short )
				, &bytesTransferred, timeout );
			if( libusbResult != LIBUSB_SUCCESS ) {
				result = LIBUSB_RESULT_TO_AIOUSB_RESULT( libusbResult );
				break;							// from while()
			} else if( bytesTransferred != ( int ) ( wordsWritten * sizeof( unsigned short ) ) ) {
				result = AIOUSB_ERROR_INVALID_DATA;
				break;							// from while()
			} else {
				bytesTransferred = libusb_control_transfer( deviceHandle
					, USB_WRITE_TO_DEVICE, AUR_LOAD_BULK_CALIBRATION_BLOCK
					, sramAddress, wordsWritten, 0, 0, timeout );
				if( bytesTransferred != 0 ) {
					result = LIBUSB_RESULT_TO_AIOUSB_RESULT( bytesTransferred );
					break;						// from while()
				}	// if( bytesTransferred ...
			}	// else if( bytesTransferred ...
			wordsRemaining -= wordsWritten;
			sramAddress += wordsWritten;
		}	// while( wordsRemaining ...
	} else {
		result = AIOUSB_ERROR_DEVICE_NOT_CONNECTED;
		AIOUSB_UnLock();
	}	// if( deviceHandle ...

	return result;
}	// AIOUSB_ADC_SetCalTable()
// }}}

// {{{ AIOUSB_ADC_InternalCal()
unsigned long AIOUSB_ADC_InternalCal(
	unsigned long DeviceIndex
	, AIOUSB_BOOL autoCal
	, unsigned short returnCalTable[]
	, const char *saveFileName
) {
	if( ! AIOUSB_Lock() )
		return AIOUSB_ERROR_INVALID_MUTEX;

	unsigned long result = AIOUSB_Validate( &DeviceIndex );
	if( result != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( result ...

	DeviceDescriptor *const deviceDesc = &deviceTable[ DeviceIndex ];
	if( deviceDesc->bADCStream == AIOUSB_FALSE ) {
		AIOUSB_UnLock();
		return AIOUSB_ERROR_NOT_SUPPORTED;
	}	// if( deviceDesc->bADCStream ...

	if( ( result = ADC_QueryCal( DeviceIndex ) ) != AIOUSB_SUCCESS ) {
		AIOUSB_UnLock();
		return result;
	}	// if( ( result ...

	AIOUSB_UnLock();
	unsigned short *const calTable = ( unsigned short * ) malloc( CAL_TABLE_WORDS * sizeof( unsigned short ) );
	assert( calTable != 0 );
	if( calTable != 0 ) {
		if( autoCal ) {
			/*
			 * create calibrated calibration table
			 */
			result = ReadConfigBlock( DeviceIndex, AIOUSB_FALSE );
			if( result == AIOUSB_SUCCESS ) {
				AIOUSB_Lock();
				const ADConfigBlock origConfigBlock = deviceDesc->cachedConfigBlock;	// restore when done
				AIOUSB_SetAllGainCodeAndDiffMode( &deviceDesc->cachedConfigBlock, AD_GAIN_CODE_0_10V, AIOUSB_FALSE );
				AIOUSB_SetTriggerMode( &deviceDesc->cachedConfigBlock, 0 );
				AIOUSB_SetScanRange( &deviceDesc->cachedConfigBlock, 0, 0 );
				AIOUSB_SetOversample( &deviceDesc->cachedConfigBlock, 0 );
				AIOUSB_UnLock();
				double averageCounts, groundCounts, referenceCounts;
				int reading;
				for( reading = 0; reading <= 1; reading++ ) {
					AIOUSB_Lock();
					AIOUSB_SetCalMode( &deviceDesc->cachedConfigBlock
						, ( reading == 0 )
							? AD_CAL_MODE_GROUND
							: AD_CAL_MODE_REFERENCE );
					AIOUSB_UnLock();			// unlock while communicating with device
					result = WriteConfigBlock( DeviceIndex );
					if( result == AIOUSB_SUCCESS ) {
						/*
						 * average a bunch of readings to get a nice, stable reading
						 */
						const int AVERAGE_SAMPLES = 256;
						const unsigned MAX_GROUND = 0x00ff
							, MIN_REFERENCE = 0xf000;
						long countsSum = 0;
						int sample;
						unsigned short counts[ MAX_IMM_ADCS ];
						for( sample = 0; sample < AVERAGE_SAMPLES; sample++ ) {
							result = ADC_GetImmediate( DeviceIndex, 0, counts );
							if( result == AIOUSB_SUCCESS )
								countsSum += counts[ 0 ];
							else
								goto abort;
						}	// for( sample ...
						averageCounts = countsSum / ( double ) AVERAGE_SAMPLES;
						if( reading == 0 ) {
							if( averageCounts <= MAX_GROUND )
								groundCounts = averageCounts;
							else {
								result = AIOUSB_ERROR_INVALID_DATA;
								goto abort;
							}	// if( averageCounts ...
						} else {
							if(
								averageCounts >= MIN_REFERENCE
								&& averageCounts <= AI_16_MAX_COUNTS
							)
								referenceCounts = averageCounts;
							else {
								result = AIOUSB_ERROR_INVALID_DATA;
								goto abort;
							}	// if( averageCounts ...
						}	// if( reading ...
					} else
						goto abort;
				}	// for( readings ...
abort:
				AIOUSB_Lock();
				deviceDesc->cachedConfigBlock = origConfigBlock;
				AIOUSB_UnLock();				// unlock while communicating with device
				WriteConfigBlock( DeviceIndex );

				if( result == AIOUSB_SUCCESS ) {
					/*
					 * we have good ground and reference readings; calculate table that makes
					 * ground reading equal to 0.0V and reference reading equal to 9.933939V;
					 * the target reference voltage of 9.933939V is equal to the target of
					 * 9.940108V provided by AP on 21-Oct-09, minus a 0.006169V offset added
					 * to all A/D readings
					 */
					const unsigned TARGET_GROUND_COUNTS = 0		// == 0.0V on 0-10V range (0.0V / 10.0V * 65535 = 0x0000)
						, TARGET_REFERENCE_COUNTS = 0xfe4e;		// == 9.933939V on 0-10V range (9.933939V / 10.0V * 65535 = 0xfe4e)
					const double slope
						= ( double ) ( TARGET_REFERENCE_COUNTS - TARGET_GROUND_COUNTS )
						/ ( referenceCounts - groundCounts );
					const double offset = ( double ) TARGET_GROUND_COUNTS - slope * groundCounts;
					int index;
					for( index = 0; index < CAL_TABLE_WORDS; index++ ) {
						long value = ( long ) round( slope * index + offset );
						if( value < 0 )
							value = 0;
						else if( value > AI_16_MAX_COUNTS )
							value = AI_16_MAX_COUNTS;
						calTable[ index ] = ( unsigned short ) value;
					}	// for( index ...
				}	// if( result
			}	// if( result ...
		} else {
			/*
			 * create default (1:1) calibration table; that is, each output word equals the input word
			 */
			int index;
			for( index = 0; index < CAL_TABLE_WORDS; index++ )
				calTable[ index ] = ( unsigned short ) index;
		}	// if( autoCal )

		if( result == AIOUSB_SUCCESS ) {
			/*
			 * optionally return calibration table to caller
			 */
			if( returnCalTable != 0 )
				memcpy( returnCalTable, calTable, CAL_TABLE_WORDS * sizeof( unsigned short ) );

			/*
			 * optionally save calibration table to a file
			 */
			if( saveFileName != 0 ) {
				FILE *const calFile = fopen( saveFileName, "w" );
				if( calFile != NULL ) {
					const size_t wordsWritten = fwrite( calTable, sizeof( unsigned short ), CAL_TABLE_WORDS, calFile );
					fclose( calFile );
					if( wordsWritten != ( size_t ) CAL_TABLE_WORDS ) {
						remove( saveFileName );		// file is likely corrupt or incomplete
						result = AIOUSB_ERROR_FILE_NOT_FOUND;
					}	// if( wordsWritten ...
				} else
					result = AIOUSB_ERROR_FILE_NOT_FOUND;
			}	// if( saveFileName ...

			/*
			 * finally, send calibration table to device
			 */
			result = AIOUSB_ADC_SetCalTable( DeviceIndex, calTable );
		}	// if( result ...

		free( calTable );
	} else
		result = AIOUSB_ERROR_NOT_ENOUGH_MEMORY;

	return result;
}	// AIOUSB_ADC_InternalCal()
// }}}

// {{{ functions for manipulating A/D configuration block

/*
 * we have to lock some of these functions because they access the device table; we don't
 * have to lock functions that don't access the device table
 */

void AIOUSB_InitConfigBlock( ADConfigBlock *config, unsigned long DeviceIndex, AIOUSB_BOOL defaults ) {
	assert( config != 0 );
	if( AIOUSB_Lock() ) {
		if( AIOUSB_Validate( &DeviceIndex ) == AIOUSB_SUCCESS ) {
			config->device = &deviceTable[ DeviceIndex ];
			config->size = ( ( DeviceDescriptor * ) config->device )->ConfigBytes;
			if( defaults ) {
				AIOUSB_SetAllGainCodeAndDiffMode( config, AD_GAIN_CODE_0_10V, AIOUSB_FALSE );
				AIOUSB_SetCalMode( config, AD_CAL_MODE_NORMAL );
				AIOUSB_SetTriggerMode( config, 0 );
				AIOUSB_SetScanRange( config, 0, ( ( DeviceDescriptor * ) config->device )->ADCMUXChannels - 1 );
				AIOUSB_SetOversample( config, 0 );
			}	// if( defaults )
		}	// if( AIOUSB_Validate( ...
		AIOUSB_UnLock();
	} else {
		// mark as uninitialized
		config->device = 0;
		config->size = 0;
	}	// if( AIOUSB_Lock() )
}	// AIOUSB_InitConfigBlock()


void AIOUSB_SetAllGainCodeAndDiffMode( ADConfigBlock *config, unsigned gainCode, AIOUSB_BOOL differentialMode ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
	) {
		gainCode &= AD_GAIN_CODE_MAX;
		if( differentialMode )
			gainCode |= AD_DIFFERENTIAL_MODE;
		unsigned channel;
		for( channel = 0; channel < AD_NUM_GAIN_CODE_REGISTERS; channel++ )
			config->registers[ AD_CONFIG_GAIN_CODE + channel ] = gainCode;
	}	// if( config->device ...
}	// AIOUSB_SetAllGainCodeAndDiffMode()


unsigned AIOUSB_GetGainCode( const ADConfigBlock *config, unsigned channel ) {
	assert( config != 0 );
	unsigned gainCode = AD_GAIN_CODE_MIN;		// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		gainCode = (
			config->registers[ AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ]
			& ( unsigned char ) AD_GAIN_CODE_MAX
		);
	}	// if( config->device ...
	return gainCode;
}	// AIOUSB_GetGainCode()


void AIOUSB_SetGainCode( ADConfigBlock *config, unsigned channel, unsigned gainCode ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
	) {
		unsigned char *const reg = &config->registers[ AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ];
		assert( reg != 0 );
		*reg
			= ( *reg & ~( unsigned char ) AD_GAIN_CODE_MAX )
			| ( unsigned char ) ( gainCode & AD_GAIN_CODE_MAX );
	}	// if( config->device ...
}	// AIOUSB_SetGainCode()


AIOUSB_BOOL AIOUSB_IsDifferentialMode( const ADConfigBlock *config, unsigned channel ) {
	assert( config != 0 );
	AIOUSB_BOOL differentialMode = AIOUSB_FALSE;
	if(
		config->device != 0
		&& config->size != 0
	) {
		differentialMode
			= (
				(
					config->registers[ AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ]
					& ( unsigned char ) AD_DIFFERENTIAL_MODE
				) != 0
			)
				? AIOUSB_TRUE
				: AIOUSB_FALSE;
	}	// if( config->device ...
	return differentialMode;
}	// AIOUSB_IsDifferentialMode()


void AIOUSB_SetDifferentialMode( ADConfigBlock *config, unsigned channel, AIOUSB_BOOL differentialMode ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
	) {
		unsigned char *const reg = &config->registers[ AD_CONFIG_GAIN_CODE + channel % AD_NUM_GAIN_CODE_REGISTERS ];
		assert( reg != 0 );
		if( differentialMode )
			*reg |= ( unsigned char ) AD_DIFFERENTIAL_MODE;
		else
			*reg &= ~( unsigned char ) AD_DIFFERENTIAL_MODE;
	}	// if( config->device ...
}	// AIOUSB_SetDifferentialMode()


unsigned AIOUSB_GetCalMode( const ADConfigBlock *config ) {
	assert( config != 0 );
	unsigned calMode = AD_CAL_MODE_NORMAL;		// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		calMode = config->registers[ AD_CONFIG_CAL_MODE ];
	}	// if( config->device ...
	return calMode;
}	// AIOUSB_GetCalMode()


void AIOUSB_SetCalMode( ADConfigBlock *config, unsigned calMode ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
		&& (
			calMode == AD_CAL_MODE_NORMAL
			|| calMode == AD_CAL_MODE_GROUND
			|| calMode == AD_CAL_MODE_REFERENCE
		)
	) {
		config->registers[ AD_CONFIG_CAL_MODE ] = ( unsigned char ) calMode;
	}	// if( config->device ...
}	// AIOUSB_SetCalMode()


unsigned AIOUSB_GetTriggerMode( const ADConfigBlock *config ) {
	assert( config != 0 );
	unsigned triggerMode = 0;					// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		triggerMode = config->registers[ AD_CONFIG_TRIG_COUNT ];
	}	// if( config->device ...
	return triggerMode;
}	// AIOUSB_GetTriggerMode()


void AIOUSB_SetTriggerMode( ADConfigBlock *config, unsigned triggerMode ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
		&& ( triggerMode & ~AD_TRIGGER_VALID_MASK ) == 0
	) {
		config->registers[ AD_CONFIG_TRIG_COUNT ] = ( unsigned char ) triggerMode;
	}	// if( config->device ...
}	// AIOUSB_SetTriggerMode()


unsigned AIOUSB_GetStartChannel( const ADConfigBlock *config ) {
	assert( config != 0 );
	unsigned startChannel = 0;					// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		if( config->size == AD_MUX_CONFIG_REGISTERS )
			startChannel
				= ( ( config->registers[ AD_CONFIG_MUX_START_END ] & 0x0f ) << 4 )
				| ( config->registers[ AD_CONFIG_START_END ] & 0xf );
		else
			startChannel = ( config->registers[ AD_CONFIG_START_END ] & 0xf );
	}	// if( config->device ...
	return startChannel;
}	// AIOUSB_GetStartChannel()


unsigned AIOUSB_GetEndChannel( const ADConfigBlock *config ) {
	assert( config != 0 );
	unsigned endChannel = 0;					// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		if( config->size == AD_MUX_CONFIG_REGISTERS )
			endChannel
				= ( config->registers[ AD_CONFIG_MUX_START_END ] & 0xf0 )
				| ( config->registers[ AD_CONFIG_START_END ] >> 4 );
		else
			endChannel = ( config->registers[ AD_CONFIG_START_END ] >> 4 );
	}	// if( config->device ...
	return endChannel;
}	// AIOUSB_GetEndChannel()


void AIOUSB_SetScanRange( ADConfigBlock *config, unsigned startChannel, unsigned endChannel ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
		&& AIOUSB_Lock()
	) {
		if( config->size == AD_MUX_CONFIG_REGISTERS ) {
			// this board has a MUX, so support more channels
			if(
				endChannel < ( ( DeviceDescriptor * ) config->device )->ADCMUXChannels
				&& startChannel <= endChannel
			) {
				config->registers[ AD_CONFIG_START_END ]
					= ( unsigned char ) ( ( endChannel << 4 ) | ( startChannel & 0x0f ) );
				config->registers[ AD_CONFIG_MUX_START_END ]
					= ( unsigned char ) ( ( endChannel & 0xf0 ) | ( ( startChannel >> 4 ) & 0x0f ) );
			}	// if( endChannel ...
		} else {
			// this board doesn't have a MUX, so support base number of channels
			if(
				endChannel < ( ( DeviceDescriptor * ) config->device )->ADCChannels
				&& startChannel <= endChannel
			) {
				config->registers[ AD_CONFIG_START_END ] = ( unsigned char ) ( ( endChannel << 4 ) | startChannel );
			}	// if( endChannel ...
		}	// if( config->size ...
		AIOUSB_UnLock();
	}	// if( config->device ...
}	// AIOUSB_SetScanRange()


unsigned AIOUSB_GetOversample( const ADConfigBlock *config ) {
	assert( config != 0 );
	unsigned overSample = 0;					// return reasonable value on error
	if(
		config->device != 0
		&& config->size != 0
	) {
		overSample = config->registers[ AD_CONFIG_OVERSAMPLE ];
	}	// if( config->device ...
	return overSample;
}	// AIOUSB_GetOversample()


void AIOUSB_SetOversample( ADConfigBlock *config, unsigned overSample ) {
	assert( config != 0 );
	if(
		config->device != 0
		&& config->size != 0
		&& overSample <= 255
	) {
		config->registers[ AD_CONFIG_OVERSAMPLE ] = ( unsigned char ) overSample;
	}	// if( config->device ...
}	// AIOUSB_SetOversample()

// }}}

// }}}

// {{{ C++ support
#ifdef __cplusplus
}	// namespace AIOUSB
#endif
// }}}


/* end of file */
